import click
import smart_open_with_pbgzip
from smart_open import smart_open
from hich.pairs import read_pairs
import polars as pl
import sys
import io

@click.command()
@click.option("--read_from", type=str, default="")
@click.option("--output_to", type=str, default="")
@click.option("--parse", type=str, nargs = 3, multiple=True, default=[], help="Format is --update [FROM_COL] [TO_COL] '[PATTERN]' as in --pattern 'readID' 'cellID' '{cellID}:{ignore}'")
@click.option("--placeholder", type=str, nargs = 2, multiple=True, default=[], help="Format is --placeholder [COL] [PLACEHOLDER] which replaces every column value with the placeholder string")
@click.option("--regex", type=str, nargs = 4, multiple=True, default=[], help="Format is --placeholder [FROM_COL] [TO_COL] [REGEX] [GROUP_INDEX] which extracts the group index specified (0=whole pattern) from the given regex from FROM_COL and sets it as the value in TO_COL")
@click.option("--drop", type=str, multiple=True, default=[], help="Column to drop")
@click.option("--select", type=str, default = "", help="Space-separated list of output column names to output in the order specified")
@click.option("--batch-size", type=int, default=10000, help="Number of records per batch")
def reshape(read_from, output_to, parse, placeholder, regex, drop, select, batch_size):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    read_from = smart_open(read_from, "rt") if read_from else sys.stdin
    reader = read_pairs(read_from, yield_columns_line=False, batch_size=batch_size)
    header = next(reader)
    output = smart_open(output_to, "wt") if output_to else None
    parse_cols = [
        (pl.col(from_col)
           .map_elements(lambda x: _parse(pattern, x)[to_col], return_dtype=pl.String)
           .alias(to_col))
        for from_col, to_col, pattern in parse
    ]
    regex_cols = [
        pl.col(from_col).str.extract(pattern, int(group_index)).alias(to_col)
        for from_col, to_col, pattern, group_index in regex
    ]
    placeholder_cols = [
        pl.lit(lit).alias(col)
        for col, lit in placeholder
    ]
    update_cols = parse_cols + regex_cols + placeholder_cols
    select = select.split()
    
    header_written = False

    if output:
        output.write(header)
    else:
        click.echo(header, nl=False)
    for df in reader:
        df = df.with_columns(*update_cols) if update_cols else df
        df = df.drop(drop) if drop else df
        df = df.select(select) if select else df
        if output:
            if not header_written:
                output.write("#columns: " + " ".join(df.columns) + "\n")
                header_written = True
            df = df.to_pandas()
            df.to_csv(output, sep="\t", header=False, index=False)
        else:
            df = df.to_pandas()
            buffer = io.StringIO()
            if not header_written:
                click.echo("#columns: " + " ".join(df.columns) + "\n", nl=False)
                header_written = True

            df.to_csv(buffer, sep="\t", header=False, index=False)
            click.echo(buffer.getvalue(), nl=False)


from dataclasses import dataclass, field
from hich.parse.pairs_segment import PairsSegment
from hich.stats.discrete_distribution import DiscreteDistribution
from polars import DataFrame
from typing import Callable, List
import time
import bisect

# Latest code review: 2024/10/17, Ben Skubi
# See tests/test_pairs_classifier.py for unit tests

def nat_partition(cuts):
    return sorted(list(set(cuts + [float('inf')]))) if cuts else None

@dataclass
class PairsClassifier:
    """Classifies PairsSegment objects into outcomes.

    The PairsClassifier defines a way of mapping PairsSegment objects into
    tuples, representing outcomes. An immersive example is:

    PairsClassifier(['record.chr1', 'record.chr2', 'stratum'], [1000, 10000])

    This PairsClassifier will categorize PairsSegment objects ('record') according
    to chr1, chr2, and strata (a binning of the distance between pos1 and pos2).
    If the distance is larger than the highest strata, 'inf' will be returned.

    This works by generating Python code expressing a tuple that combines
    the conjuncts 'record.chr1', 'record.chr2', and 'stratum'. This is compiled
    and used to obtain the relevant data as a tuple. The resulting tuple is
    the return value of the 'classify' function.
    """
    conjuncts: list[str] = field(default_factory=list)
    cis_strata: list = field(default_factory=list)
    classification_code: str = ""

    def __post_init__(self) -> None:
        """Creates and compiles Python classification code that converts conjuncts into a tuple"""
        # Generate Python code for conjuncts that can be conveniently read by eval()
        # to classify PairsSegments
        self.classification_code = self.generate_classification_code()

        # Compile to Python bytecode
        self.compiled_classification_code = compile(self.classification_code, '<string>', 'eval') if self.classification_code else None

        # If strata is given, add 'inf' to it to clearly define the largest stratum
        self.cis_strata = nat_partition(self.cis_strata)

    def generate_classification_code(self):
        """Preprents 'record.' to conjuncts that need it, then format as Python code tuple or "" if empty"""
        # Create a comma-separated list of the conjuncts and add parenthesis
        def needs_prefix(conjunct):
            conjunct_has_record_prefix = conjunct.strip().startswith("record.")
            conjunct_needs_no_record_prefix = conjunct in ["strata", "stratum"]
            conjunct_needs_prefix = not (conjunct_has_record_prefix or conjunct_needs_no_record_prefix)
            return conjunct_needs_prefix
        
        # Prepend with "record." unless conjunct is "stratum" or "strata"
        def add_prefix_if_necessary(conjunct):
            return f"record.{conjunct.strip()}" if needs_prefix(conjunct) else conjunct

        # Format the conjuncts as a tuple (conjuncts[0], conjuncts[1], ...) unless empty
        # in which case return ""
        formatted_conjuncts = [add_prefix_if_necessary(conjunct) for conjunct in self.conjuncts]
        tuple_center = ", ".join(formatted_conjuncts) if formatted_conjuncts else None
        return "(" + tuple_center + ",)" if tuple_center else ""


    def get_stratum(self, pair: PairsSegment) -> int | None:
        """Categorize PairsSegment into stratum
        
            If strata are defined, returns:
                If pair.is_is(): the smallest stratum value S such that pair.distance <= S
                Otherwise, an empty string
            If strata not defined, returns None
        """
        if self.cis_strata and pair.is_cis:
            index = bisect.bisect_left(self.cis_strata, pair.distance)
            result = self.cis_strata[index]
            return result
        else:
            return None

    def classify(self, record: PairsSegment) -> tuple:
        """Generate a tuple with the classification code.
        
        The code can access the PairsSegment 'record', the stratum to which it
        maps 'stratum' (an integer if it is cis and strata are defined, an empty string
        if it is not cis and strata are defined, None if strata are undefined),
        and 'strata' (a list of integers or None).
        """
        
        if not self.compiled_classification_code:
            message = "No compiled classification code"
            if not self.conjuncts:
                message += "and list of conjuncts is empty."
            else:
                message += f"but list of conjuncts is {self.conjuncts}"
            raise TypeError(message)
        if not isinstance(record, PairsSegment):
            raise TypeError(f"Expected object of type PairsSegment, but got {type(record)}")
        
        
        strata = self.cis_strata
        stratum = self.get_stratum(record)
        result = eval(self.compiled_classification_code)
        return result
    
    def to_polars(self, distribution: DiscreteDistribution) -> DataFrame:
        """Output conjuncts plus 'count' as columns, rows as events + observed count
        
        Note: doesn't output the strata.
        """
        schema = self.conjuncts + ["count"]
        rows = [event + (count,) for event, count in distribution.items()]
        return DataFrame(rows, schema=schema, orient='row')
    
    def from_polars(self, df: DataFrame, cis_strata: List[int] = None) -> DiscreteDistribution:
        """Treats all columns except 'count' as conjuncts, the 'count' column as the counts for each outcome
        """
        self.cis_strata = cis_strata
        self.conjuncts = [col for col in df.columns if col != "count"]
        self.__post_init__()
        distribution = DiscreteDistribution()
        for row in df.iter_rows():
            event = row[:-1]
            count = row[-1]
            distribution[event] = count
        return distribution

    def __getstate__(self):
        # Return a dictionary excluding the compiled code
        state = self.__dict__.copy()
        del state['compiled_classification_code']
        return state

    def __setstate__(self, state):
        # Restore the object's state and recompile the code
        self.__dict__.update(state)
        self.compiled_classification_code = compile(self.classification_code, '<string>', 'eval')




import polars as pl

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

class BedpePairs:
    def __init__(self, df):
        self.df = df
    
    def fragtag(self, frag_index):
        """Map .pairs end pos to restriction fragment index, start, end
        
        This algo centers around numpy searchsorted which will "Find indices
        where elements should be inserted to maintain order." We partition each
        end by chromosome, sort by position, then apply np.searchsorted on
        the sorted fragment index ends. This gives the fragment each end
        maps to, and from this we can extract the start and end position of the
        fragment.

        This mapping depends on partitioning each pair by chromosome, then by
        position. To make it possible to restore order, we start by assigning
        a row index to the pairs, which is extracted along with the chrom and
        pos for each end of the pair.
        """

        # Assign pair id to restore order after mapping pair ends.
        pair_id = "__pairID__"
        pair_id_df = self.df.with_row_index(pair_id)
        

        for end in ["1", "2"]:
            # Get end-specific column names
            chrom_col = f"chrom{end}"
            pos_col = f"pos{end}"
            frag_colnames = [f"rfrag{end}", f"rfrag_start{end}", f"rfrag_end{end}"]

            # Select the pair id, chromosome, and position, then split into a 
            # {chromosome : dataframe} dictionary to facilitate per-chromosome
            # fragment mapping.
            end_chroms = pair_id_df.select(pair_id, chrom_col, pos_col) \
                                   .partition_by([chrom_col], as_dict = True)

            for chrom, chrom_df in end_chroms.items():
                # Iterate through the chromosomes and chrom- and pair end-
                # specific dataframes.

                # Ensure dataframe is sorted by position to allow searchsorted
                # Then extract positions and map them to restriction fragments
                chrom_df = chrom_df.sort(by = [pos_col])
                positions = chrom_df[pos_col].to_list()
                frag_cols = BedpePairs.frag_columns(frag_index,
                                                    chrom,
                                                    positions,
                                                    frag_colnames)
                
                # Add frag columns to the dataframe subset used for mapping.
                # As we'll be recovering the original row order by sorting by
                # pair ID and the chrom and pos are preserved in the original
                # dataframe, we drop the chrom and pos from the fragments
                # dataframe. Then we extract just the fragment column names
                # once they are sorted by pair ID and attach them to the 
                # complete per-chromosome dataframe.
                chrom_df = chrom_df.with_columns(frag_cols)
                chrom_df = chrom_df.drop([chrom_col, pos_col])
                frag_cols = chrom_df.sort(by = [pair_id]) \
                                    .select(frag_colnames)                              
                end_chroms[chrom] = end_chroms[chrom].with_columns(frag_cols)
            
            # Reconcatenate all the per-chrom dataframes with the fragment
            # columns and drop (I believe the drop is redundant but haven't
            # checked.)
            end_frags = pl.concat(end_chroms.values()) \
                          .drop([chrom_col, pos_col])

            # Join on the pair id
            pair_id_df = pair_id_df.join(end_frags,
                                         on = [pair_id],
                                         how = "inner",
                                         coalesce = True)
        # Resort by the original pair id to restore the original order, then
        # drop the pair id column to get the original dataframe with the
        # fragment columns for end 1 and end 2 as the final 6 columns.
        pair_id_df = pair_id_df.sort(by = [pair_id])
        return pair_id_df.drop(pair_id)

    @classmethod
    def frag_columns(self,
                         frag_index,
                         chrom,
                         positions,
                         frag_colnames):
        if chrom not in frag_index:
            count = len(positions)
            rfrag = pl.repeat(-1, count)
            rfrag_start = pl.repeat(0, count)
            rfrag_end = pl.repeat(0, count)
        else:
            # Get the indices where the positions sort to in the restriction fragments
            # for the chromosome -- this is the main event that identifies the restriction fragments
            rfrag_indices = frag_index.search(chrom, positions)

            # Create column for the restriction fragment index and for the start and end positions
            rfrag = pl.Series(rfrag_indices)

            # Create column for start position of the restriction fragment
            # Adjust exact positioning to match pairtools output
            # (Set 0 to -1, then add 1 to all)
            rfrag_start = frag_index.starts(chrom) \
                            .gather(rfrag_indices) \
                            .replace(0, -1) \
                            + 1

            # Create column for end position of the restriction fragment
            # Adjust exact position to match pairtools output (add 1)
            rfrag_end = frag_index.ends(chrom) \
                        .gather(rfrag_indices) \
                        + 1
        
        # Cast and set the name of the fragment columns
        frag_cols = [rfrag, rfrag_start, rfrag_end]
        for i, frag_col in enumerate(frag_cols):
            frag_cols[i] = frag_col.cast(pl.Int64).alias(frag_colnames[i])

        return frag_cols


import click
from hich.commands import *

@click.group
def hich():
    pass

@hich.group 
def view(): pass

hich.add_command(compartments)
hich.add_command(create_scool)
hich.add_command(downsample)
hich.add_command(digest)
hich.add_command(fragtag)
hich.add_command(gather)
#hich.add_command(organize)
hich.add_command(hicrep)
hich.add_command(reshape)
hich.add_command(stats)
hich.add_command(stats_aggregate)
view.add_command(hicrep_comparisons)

if __name__ == "__main__":
    hich()



import click
from hich.visuals import view_hicrep

@click.command(name='hicrep')
@click.option("--host", default = "127.0.0.1", show_default = True)
@click.option("--port", default = 8050, show_default = True)
def hicrep_comparisons(host, port):
    view_hicrep.run_dashboard(host, port)


import click
from Bio import SeqIO
from Bio.Restriction import RestrictionBatch
from Bio.Seq import Seq
import sys
from itertools import chain
from smart_open import smart_open
import polars as pl

def sorted_unique_cut_sites(seq_record, enzymes, cutshift = 1):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    df = {"chrom":[], "start":[], "end":[]}
    digest = enzymes.search(seq_record.seq)
    digest = [digest[enzyme] for enzyme in digest]
    cut_sites = sorted(list(set(chain(*digest))))
    if cut_sites:
        cut_sites = (pl.Series(cut_sites) + cutshift).to_list()
    return cut_sites

def chrom_frags_df(seq_record, enzymes, cutshift = 1):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    start = [0]
    end = [len(seq_record.seq)]
    frag_ends = start + sorted_unique_cut_sites(seq_record, enzymes, cutshift) + end
    
    chrom = seq_record.id
    frag_count = len(frag_ends)-1

    chrom_col = [chrom]*frag_count
    start_col = frag_ends[:-1]
    end_col = frag_ends[1:]
    df = {"chrom":chrom_col, "start":start_col, "end":end_col}
    return pl.DataFrame(df)

def write_bed_file(frag_index, output_file):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    handle = smart_open(output_file, "w") if output_file else sys.stdout
    frag_index.write_csv(handle, include_header=False, separator="\t")

def make_frag_index(reference_filename,
                    enzyme_names,
                    output_file,
                    startshift = 0,
                    endshift = 0,
                    cutshift = 1):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    # Load the reference genome
    reference_file = smart_open(reference_filename, "rt")
    seq_record = SeqIO.parse(reference_file, "fasta")
    genome = SeqIO.to_dict(seq_record)

    enzymes = RestrictionBatch(enzyme_names)
    
    # Process each sequence record in the genome
    chrom_frag_indexes = []
    for seq_record in genome.values():
        chrom_frag_index = chrom_frags_df(seq_record, enzymes, cutshift)
        chrom_frag_indexes.append(chrom_frag_index)

    frag_index = pl.concat(chrom_frag_indexes)
    frag_index = frag_index.with_columns(pl.col('start') + startshift)
    frag_index = frag_index.with_columns(pl.col('end') + endshift)

    # Write to BED file
    write_bed_file(frag_index, output_file)

def kit_names_to_enzymes(digest):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    digest = set(digest)
    enzymes = set()

    # If the digest is a known kit name, map it to the corresponding enzymes
    kits = {
        ("Arima Genome-Wide HiC+", "Arima"): ["DpnII", "HinfI"],
        ("Phase Proximo 2021+ Plant", "Phase Plant"): ["DpnII"],
        ("Phase Proximo 2021+ Animal", "Phase Animal"): ["DpnII"],
        ("Phase Proximo 2021+ Microbiome", "Phase Microbiome"): ["Sau3AI", "MluCI"],
        ("Phase Proximo 2021+ Human", "Phase Human"): ["DpnII"],
        ("Phase Proximo 2021+ Fungal", "Phase Fungal"): ["DpnII"]
    }
    digest_removed = set()
    for kit, kit_enzymes in kits.items():
        for digest_item in digest:
            if digest_item in kit:
                for enzyme in kit_enzymes:
                    enzymes.add(enzyme)
                digest_removed.add(digest_item)
    
    digest = digest.difference(digest_removed)

    # Add any non-kit enzymes to the digest
    enzymes = enzymes.union(digest)

    print(enzymes)
    return enzymes

def make_fragment_index(output, startshift, endshift, cutshift, reference, digest):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    enzyme_names = kit_names_to_enzymes(digest)

    return make_frag_index(reference,
                    enzyme_names,
                    output,
                    startshift,
                    endshift,
                    cutshift)


"""Run hicrep on combinations of files and parameters"""

import click
import numpy as np
from hicrep.utils import readMcool
from .hicrep_wrapper import hicrepSCC
import os
import glob
import cooler
from cooler import Cooler
from dataclasses import *
import h5py

import warnings
from pathlib import Path
from typing import Tuple, List, Callable
from collections import defaultdict
import polars as pl
from itertools import combinations, combinations_with_replacement, product, chain

from collections.abc import Iterable
import concurrent.futures
warnings.simplefilter(action='ignore', category=FutureWarning)

"""
Get combinations of hicrep comparisons from the user, generate, and return or
save a dataframe specifying the parameters and results.

1. Get combinations of params and matrices
2. Store in polars dataframe
3. Compute SCC and distance
4. Save as tsv

"""

class SCC(float):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    def distance(self):
        (.5*(1-self))**.5 if self != np.nan else 0

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class HicrepCall:
    file1: object
    file2: object
    resolution: int
    h: int
    dBPMax: int
    bDownSample: bool
    chrom: str
    scc: SCC = None

    def cools(self):
        cool1, _ = readMcool(self.file1, self.resolution)
        cool2, _ = readMcool(self.file2, self.resolution)
        return (cool1, cool2)

    def run_hicrep(self):        
        self.scc = SCC(hicrepSCC(*self.cools(),
                                 self.h,
                                 self.dBPMax,
                                 self.bDownSample,
                                 [self.chrom]))
        return self

def run_hicrep(call):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    return call.run_hicrep()

def parallel_hicrep(callers, max_workers = None):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    with concurrent.futures.ProcessPoolExecutor(max_workers = max_workers) as executor:
        results = executor.map(run_hicrep, callers)
    if results:
        columns = list(HicrepCall.__dataclass_fields__.keys())
        rows = [astuple(result) for result in results]
        return pl.DataFrame(rows, orient='row', schema=columns)

def shared_chroms(filenames: List[str], filter = lambda chrom, size: chrom):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    all_chroms = None
    for filename in filenames:
        coolers = []
        if cooler.fileops.is_cooler(filename):
            coolers = [Cooler(filename)]
        elif cooler.fileops.is_multires_file(filename):
            coolers = [Cooler(f"{filename}::{cooler}") for cooler in cooler.fileops.list_coolers(filename)]
        else:
            raise Exception(f"shared_chroms received filename {filename} that is not cooler or multires cooler file")
        
        coolers_chromsizes = {}
        for c in coolers:
            cooler_chroms = set()
            for chrom, size in c.chromsizes.to_dict().items():
                result = filter(chrom, size)
                if result:
                    cooler_chroms.add(result)
            all_chroms = all_chroms.intersection(cooler_chroms) if all_chroms else cooler_chroms

    return all_chroms

def hicrep_callers(matrices: List,
                   resolutions: List[int],
                   chroms: List[str],
                   h: List[int],
                   dBPMax: List[int],
                   bDownSample: List[bool],
                   matrix_pair_function = combinations_with_replacement,
                   param_set_function = product):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    
    def asiterable(arg):
        return arg if isinstance(arg, Iterable) else [arg]
    
    matrices = asiterable(matrices)
    resolutions = asiterable(resolutions)
    h = asiterable(h)
    dBPMax = asiterable(dBPMax)
    bDownSample = asiterable(bDownSample)
    chroms = asiterable(chroms)

    matrix_pairs = matrix_pair_function(matrices, 2)
    param_set = param_set_function(resolutions, h, dBPMax, bDownSample, chroms)
    combos = list(product(matrix_pairs, param_set))
    
    return [HicrepCall(combo[0][0], combo[0][1], *combo[1]) for combo in combos]
    
def hicrep_combos(resolutions, chroms, exclude, chromFilter, h, d_bp_max, b_downsample, nproc, output, paths):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    assert paths, "No paths specified in hich hicrep"
    chroms = chroms or shared_chroms(paths, lambda chrom, size: eval(chromFilter))
    chroms = set(chroms) - set(exclude) if exclude else chroms
    assert chroms, "No chromosomes specified or no universally overlapping chromosomes found in hich hicrep"
    callers = hicrep_callers(paths, resolutions, chroms, h = h, dBPMax = d_bp_max, bDownSample = b_downsample)
    result = parallel_hicrep(callers, max_workers = nproc)
    if output:
        result.write_csv(output, separator = "\t")
    else:
        return result



from hich.parse.file_splitter import FileSplitter
from dataclasses import dataclass, field
import pysam
import subprocess

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class AlignmentSplitter(FileSplitter):
    header: pysam.AlignmentHeader = None
    samtools_flags: str = None

    def access(self, filename: str):
        """Create/retrieve a .sam or .bam file write handle with header"""
        if filename not in self.handles:
            if self.samtools_flags is None:
                self.samtools_flags = "-b" if filename.endswith(".bam") else ""
            self.handles[filename] = subprocess.Popen(
                ["samtools", "view", self.samtools_flags, "-o", filename],
                stdin=subprocess.PIPE,
                text=True
            )
            if self.header: self.handles[filename].stdin.write(str(self.header))

        return self.handles[filename].stdin

    def write(self, filename: str, record: pysam.AlignedSegment):
        """Write a pysam AlignedSegment record to the given file"""
        self.access(filename).write(record.to_string() + "\n")


from hich.parse.file_splitter import FileSplitter
from dataclasses import dataclass, field
from Bio import SeqIO
import io
import subprocess
from smart_open import smart_open

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class SeqIOSplitter(FileSplitter):
    format_name: str = None

    def access(self, filename: str):
        if filename not in self.handles:
            self.handles[filename] = smart_open(filename, "w")
        return self.handles[filename]
    
    def write(self, filename: str, record: SeqIO.SeqRecord):
        output = io.StringIO()
        SeqIO.write(record, output, self.format_name)
        formatted_string = output.getvalue()
        output.close()
        self.access(filename).write(formatted_string)



from dataclasses import dataclass
import sys

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class SamheaderFragtag:
    AT: str = "HC"
    ID: str = "hich_fragtag"
    PN: str = "hich_fragtag"
    CL: str = " ".join(sys.argv)
    PP: str = ""
    VN: str = "0.1"

    def __str__(self):
        return "\t".join([
            f"#samheader: @{self.AT} ID:{self.ID}",
            f"PN:{self.PN}",
            f"CL:{self.CL}",
            f"PP:{self.PP}",
            f"VN:{self.VN}"
        ]) + "\n"


from pydantic import BaseModel, conlist, field_validator
from typing import Set, List
from collections import Counter
from hich.parse.pairs_schema import PairsSchema
from typing import Union
import warnings

class PairsColumns(BaseModel):
    columns: conlist(str, min_length=4)
    
    @field_validator('columns')
    def has_required_columns(cls, column_names: List[str]):
        unofficial = [name for name in column_names if (name in PairsSchema.common_unofficial_colnames())]
        if unofficial:
            recommended = [PairsSchema.make_official(name) for name in unofficial]
            warnings.warn(f"Attempted to create PairsColumns with common unofficial column names: {unofficial}; consider using {recommended}", UserWarning)
        
        column_names = [PairsSchema.make_official(name) for name in column_names]

        required_columns = {"chr1", "pos1", "chr2", "pos2"}
        if not required_columns.issubset(column_names):
            missing = required_columns - set(column_names)
            raise ValueError(f"Pairs file missing required columns: {', '.join(missing)}, check '#columns:' line")
        return column_names
    
    @field_validator('columns')
    def all_column_names_unique(cls, column_names: List[str]):
        column_name_counts = Counter(column_names)
        duplicates = set([name for name in column_name_counts if column_name_counts[name] > 1])
        if duplicates:
            raise ValueError(f"Attempted to create PairsColumns with duplicate column names: {duplicates}")
        return column_names

    @classmethod
    def is_columns_line(cls, line: str) -> bool:
        return line.startswith("#columns:")

    @classmethod
    def from_columns_line(cls, line: str) -> "PairsColumns":
        def crash_value_error(explain: str):
            preamble = f"Tried to create PairsColumns from '{line}', but it could not be parsed"
            raise ValueError(f"{preamble} {explain}".strip())
        
        if not PairsColumns.is_columns_line(line):
            crash_value_error("because it does not start with '#columns:'")

        fields = line.split()

        if len(fields) < 2:
            crash_value_error("because it did not contain any fields, just a '#columns:' prefix")
        
        column_names = fields[1:]
        official_column_names = [PairsSchema.make_official(name) for name in column_names]

        return PairsColumns(columns = official_column_names)

    def index(self, column: str) -> int:
        official_name = PairsSchema.make_official(column)
        return columns.index(official_name)
    
    def __getitem__(self, get: Union[int, str]) -> str:
        if isinstance(get, str):
            return self.index(get)
        return self.columns[index]
    
    def __setitem__(self, index: int, column_name: str) -> None:
        self.columns[index] = PairsSchema.make_official(column_name)
    
    def __eq__(self, other: Union[List[str], "PairsColumns"]) -> bool:
        if hasattr(other, "columns"):
            return self.columns == other.columns
        else:
            return self.columns == other


from Bio import SeqIO
import pyBigWig
import itertools
import numpy as np
import cooler
import hicstraw
from dataclasses import dataclass, field
from pathlib import Path
from cooltools.api.eigdecomp import cis_eig
from typing import List, Tuple, Dict, TextIO, TYPE_CHECKING
from scipy.sparse import coo_matrix
import polars as pl
import os
import inspect
from scipy.sparse import coo_matrix
import copy
from smart_open import smart_open

def chromsizes_hic(path: Path) -> Dict[str, int]:
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    h = hicstraw.HiCFile(str(path.resolve()))
    chroms = h.getChromosomes()
    chromsizes = {}
    for c in chroms:
        chromsizes[c.name] = c.length
    return chromsizes

def chromsizes_mcool(path: Path, resolution: int) -> Dict[str, int]:
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    abs_path = str(path.resolve())
    cooler_collections = cooler.fileops.list_coolers(abs_path)
    cooler_collection = f"/resolutions/{resolution}"
    assert cooler.fileops.is_multires_file(abs_path), f"{abs_path} is not a cooler multires file"
    assert cooler_collection in cooler_collections, f"{cooler_collection} is not a data collection in {abs_path}. Available collections are {cooler_collections}."
    c = cooler.Cooler(f"{abs_path}::{cooler_collection}")
    return c.chromsizes.to_dict()

def chromsizes(filename: Path, resolution: int = None) -> Dict[str, int]:
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    path = Path(filename)
    assert path.exists(), f"{filename} not found at {path.resolve()}"
    
    suffix = Path(filename).suffix
    if suffix == ".mcool":
        return chromsizes_mcool(path, resolution)
    elif suffix == ".hic":
        return chromsizes_hic(path)
    raise Exception(f"Extension {suffix} not supported by hich compartments called on {path.resolve()}")

def flex_chromname(real_chromnames: List[str], chromname: str, flexible: List[str] = ["chr"]):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    for real_chromname, flex in itertools.product(real_chromnames, flexible):
        if chromname == real_chromname or chromname.replace(flex, "") == real_chromname.replace(flex, ""):
            return real_chromname
    return None

def dense_cis_count_matrix_hic(path: Path, resolution: int, chrom: str):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    h = hicstraw.HiCFile(str(path.resolve()))
    resolutions = h.getResolutions()
    chromsizes = chromsizes_hic(path)
    assert resolution in resolutions, f"Resolution {resolution} not found in {path.resolve()}. Available options are: {resolutions}"
    assert chrom in chromsizes, f"Chromosome {chrom} not found in {path.resolve()}. Available options are: {list(chromsizes.keys())}"
    
    start = 0
    end = chromsizes[chrom]
    mzd = h.getMatrixZoomData(chrom, chrom, "observed", "NONE", "BP", resolution)
    
    records = mzd.getRecords(start, end, start, end)
    row = []
    col = []
    count = []
    for r in records:
        row.append(r.binY)
        col.append(r.binX)
        count.append(r.counts)
        row.append(r.binX)
        col.append(r.binY)
        count.append(r.counts)
    row = np.array(row) // resolution
    col = np.array(col) // resolution
    return coo_matrix((count, (row, col))).toarray()
                 
def dense_cis_count_matrix_mcool(path: Path, resolution: int, chrom: str):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    abs_path = str(path.resolve())
    cooler_collections = cooler.fileops.list_coolers(abs_path)
    cooler_collection = f"/resolutions/{resolution}"
    assert cooler.fileops.is_multires_file(abs_path), f"{abs_path} is not a cooler multires file"
    assert cooler_collection in cooler_collections, f"{cooler_collection} is not a data collection in {abs_path}. Available collections are {cooler_collections}."
    c = cooler.Cooler(f"{abs_path}::{cooler_collection}")
    return c.matrix(sparse=False, balance=False).fetch(chrom)

def dense_cis_count_matrix(filename: str, resolution: int, chrom: str, flex: bool = True):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    path = Path(filename)
    assert path.exists(), f"{filename} not found at {path.resolve()}"
    
    suffix = Path(filename).suffix
    if suffix == ".mcool":
        chromnames = chromsizes_mcool(filename, resolution).keys()
        chrom = flex_chromname(chromnames, chrom) if flex else chrom
        assert chrom, f"Chrom {chrom} not found in {path}"
        return dense_cis_count_matrix_mcool(filename, resolution, chrom)
    elif suffix == ".hic":
        chromnames = chromsizes_hic(filename).keys()
        chrom = flex_chromname(chromnames, chrom) if flex else chrom
        assert chrom, f"Chrom {chrom} not found in {path}"
        return dense_cis_count_matrix_hic(filename, resolution, chrom)
    raise Exception(f"Extension {suffix} not supported by hich compartments called on {path.resolve()}")

def corr_neg(a, b):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    def ok(v):
        return not np.isnan(v) and v is not None

    keep_a = []
    keep_b = []
    for ai, bi in zip(a, b):
        if ok(ai) and ok(bi):
            keep_a.append(ai)
            keep_b.append(bi)
    
    return np.corrcoef(keep_a, keep_b)[0][1] < 0

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi
@dataclass
class BEDSignal:
    chrom: str = ""
    end: int = None
    signal: list[float] = field(default_factory = list)
    starts: list[int] = field(default_factory = list)
    ends: list[int] = field(default_factory = list)
    
    def bigwig_header(self) -> Tuple[str, int]:
        return (self.chrom, self.end)

    def add_to_bigwig(self, bw):
        assert isinstance(self.chrom, str), f"BEDSignal chrom '{self.chrom}' should be str, is {type(self.chrom)}"
        assert self.lengths_match(), f"Lengths do not match for {self}"
        
        chroms = [self.chrom]*len(self.signal)
        starts = self.starts if isinstance(self.starts, list) else self.starts.tolist()
        ends = self.ends if isinstance(self.ends, list) else self.ends.tolist()
        signal = self.signal if isinstance(self.signal, list) else self.signal.tolist()
        signal = [s if not np.isnan(s) else 0.0 for s in signal]
        for it in [chroms, starts, ends, signal]:
            print(type(it), len(it), type(it[0]), all([isinstance(i, str) or not np.isnan(i) for i in it]))
        bw.addEntries(chroms, starts, ends = ends, values = signal)

    def direct(self, vec: List[float]):
        assert len(vec) == len(self.signal), f"Signals must be same ength but were {len(vec)} and {len(self.signal)}"
        vec = vec * -1 if corr_neg(self.signal, vec) else vec
        return BEDSignal(self.chrom, self.end, vec, self.starts, self.ends)

    def lengths_match(self) -> bool:
        return len(self.signal) == len(self.starts) and len(self.signal) == len(self.ends)

    def window(self, window_size: int, endpoint: int):
        self.starts = np.arange(0, endpoint, window_size)
        self.ends = np.append((self.starts[1:] - 1), endpoint)

def gc_dist(bed, seq_iter, resolution):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    while batch := tuple(itertools.islice(seq_iter, None, resolution)):
        frac_g = (batch.count("G") + batch.count("G"))/len(batch)
        g_dist.append(frac_g)
        end = start + len(batch)
        starts.append(start)
        ends.append(end)
        start = ends[-1] + 1
    return GCDist(g_dist, starts, ends)

def frac_gc_bed(seqio_record, resolution):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    seq_iter = iter(seqio_record.seq)
    chrom = seqio_record.id
    chromsize = len(seqio_record.seq)
    
    bed = BEDSignal(chrom = chrom, end = chromsize)
    bed.window(resolution, chromsize)
    while chrom_window := tuple(itertools.islice(seq_iter, None, resolution)):
        gc_count = chrom_window.count("G") + chrom_window.count("G")
        window_size_bp = len(chrom_window)
        frac_g = gc_count / window_size_bp
        bed.signal.append(frac_g)

    assert bed.lengths_match(), f"In {chrom} at resolution {resolution}, %GC BED start, end and signal vectors do not have matching length. {bed}"
    return bed

def compartment_scores(matrix: Path, resolution: int, chrom: str, guide: BEDSignal, n_eigs: int) -> List[BEDSignal]:
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    mx = dense_cis_count_matrix(matrix, resolution, chrom)
    vals, vecs = cis_eig(mx, n_eigs = n_eigs)
    return [guide.direct(vec) for vec in vecs]

def fasta_chromsizes(fasta: TextIO, chroms: List[str], exclude_chroms: List[str], keep_chroms_rule: List[str]) -> List[Tuple[str, int]]:
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    sizes = []
    todo = copy.deepcopy(chroms)
    for record in SeqIO.parse(fasta, "fasta"):
        chrom = record.id
        if chroms is not None and chrom not in chroms:
            continue
        if exclude_chroms is not None and chrom in exclude_chroms:
            chroms.remove()
            continue
        if isinstance(keep_chroms_rule, str) and not eval(keep_chroms_rule):
            if todo:
                todo.remove(chrom)
            continue
        chromsize = (record.id, len(record.seq))
        sizes.append(chromsize)

        if chroms is not None:
            todo.remove(record.id)
        if chroms and not todo:
            break
    return sizes

def write_compartment_scores(bigwig_prefix: str,
                             matrix: Path,
                             reference: Path,
                             resolution: int,
                             chroms: List[str] = None,
                             exclude_chroms: List[str] = None,
                             keep_chroms_rule: str = None,
                             n_eigs: int = 3):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    ref_abs_path = str(reference.resolve())
    ref_handle = smart_open(ref_abs_path, mode = "rt")
    bw_header = fasta_chromsizes(ref_handle, chroms, exclude_chroms, keep_chroms_rule)

    bw_filenames = [f"{bigwig_prefix}_{i}.bw" for i in range(n_eigs)]
    bigwigs = [pyBigWig.open(filename, "w") for i, filename in enumerate(bw_filenames)]
    print(bw_header)
    for bw in bigwigs:
        bw.addHeader(bw_header)
    
    ref_handle.seek(0)
    todo = copy.deepcopy(chroms)
    for record in SeqIO.parse(ref_handle, "fasta"):


        chrom = record.id
        size = len(record.seq)
        
        if chroms is not None and chrom not in chroms:
            continue
        if exclude_chroms is not None and chrom in exclude_chroms:
            chroms.remove()
            continue
        if isinstance(keep_chroms_rule, str) and not eval(keep_chroms_rule):
            if todo:
                todo.remove(chrom)
            continue

        gc = frac_gc_bed(record, resolution)
        scores = compartment_scores(matrix, resolution, chrom, gc, n_eigs)
        for score, bw in zip(scores, bigwigs):
            score.add_to_bigwig(bw)
        
        if chroms and todo:
            todo.remove(chrom)
        
        if not todo:
            break

    ref_handle.close()
    for bw in bigwigs:
        print("Closing time...", type(bw))
        bw.close()
    
    print("Opening bigwigs for testing")
    bigwigs = [pyBigWig.open(filename) for i, filename in enumerate(bw_filenames)]
    print("Displaying headers")
    for bw in bigwigs:
        print("Wrote", bw.header())




from hich.stats.discrete_distribution import DiscreteDistribution
from dataclasses import dataclass, field
from random import random

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class SelectionSampler:
    full: DiscreteDistribution = field(default_factory = DiscreteDistribution)
    target: DiscreteDistribution = field(default_factory = DiscreteDistribution)
    viewed: DiscreteDistribution = field(default_factory = DiscreteDistribution)
    kept: DiscreteDistribution = field(default_factory = DiscreteDistribution)

    def count(self, event):
        self.full[event] += 1

    def sample(self, event):
        to_view = self.full[event] - self.viewed[event]
        to_sample = self.target[event] - self.kept[event]
        keep = to_view * random() < to_sample
        self.kept[event] += keep
        self.viewed[event] += 1
        return keep

    



from hich.parse.pairs_file import PairsFile
from hich.parse.pairs_header import PairsHeader
from hich.parse.pairs_parser import PairsParser, read_pairs
from hich.parse.pairs_segment import PairsSegment
from hich.parse.pairs_splitter import PairsSplitter
from hich.stats.pairs_classifier import PairsClassifier
from hich.stats.compute_pairs_stats import compute_pairs_stats_on_path, compute_pairs_stats_on_path_list

# Facilitates importing anything pairs-related

__all__ = ['PairsFile', 'PairsHeader', 'PairsParser', 'read_pairs', 'PairsSegment', 'PairsSplitter', 'PairsClassifier',
           'compute_pairs_stats_on_path', 'compute_pairs_stats_on_path_list']


import click
from abc import ABC, abstractmethod
from typing import List

# See tests/test_cli_types.py for unit tests

class ParamList(click.ParamType, ABC):

    def __init__(self, separator = ",", do_strip = True, chars_to_strip = None):
        assert separator, f"In {str(self)}, separator was '{separator}' which will be nonfunctional. Common functional separators include delimiters like ',', ' ', ';', etc."
        self.separator = separator
        self.do_strip = do_strip
        self.chars_to_strip = chars_to_strip

    def convert(self, value, param, ctx):
        try:
            if not isinstance(value, str) and isinstance(value, self.value_type()):
                result = [value]
            else:
                result = []
                split = self.split(value)
                for item in split:
                    stripped = self.strip(item)
                    casted = self.to_type(stripped)
                    result.append(casted)

            result = [r for r in result if r is not None]
            return result
        except ValueError as e:
            self.fail(f"Invalid {self.value_type()} value in argument {param} with value {value}: {e}")
        except TypeError as e:
            self.fail(f"Using {str(self)} to parse argument '{param}' with value '{value}' ({type(value)}) resulted in error: {e}")

    def split(self, s: str) -> List[str]:
        "Split on my separator and drop empty strings from the split".strip()
        return [it for it in s.split(self.separator) if it]

    def strip(self, s: str) -> str:
        "Strip specific characters from string"
        if self.do_strip:
            return s.strip(self.chars_to_strip) if self.chars_to_strip else s.strip()
        return s

    @abstractmethod
    def value_type(self):
        pass

    @abstractmethod
    def to_type(self, value):
        pass

    def __str__(self):
        return f"{type(self)}: separator: '{self.separator}' ({type(self.separator)}), strip = '{self.do_strip}' ({type(self.do_strip)}), strip_chars: '{self.chars_to_strip}' ({type(self.chars_to_strip)})"
    


from dataclasses import dataclass, field
from smart_open import smart_open
from hich.parse.pairs_header import PairsHeader
from hich.parse.pairs_segment import PairsSegment
from pathlib import PurePath, Path

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class PairsFile:
    filepath_or_object: str = None
    mode: str = None
    header: PairsHeader = field(default_factory = PairsHeader)

    def __init__(self, filepath_or_object, mode = "rt", header = None):
        self.open(filepath_or_object, mode, header)
        self.duration = 0

    def __del__(self):
        self.close()

    def close(self):
        if hasattr(self.filepath_or_object, "close"):
            self.filepath_or_object.close()

    def open(self, filepath_or_object, mode = "r", header = None):
        # Close previously open file if necessary and set read/write/append mode
        self.close()
        
        self.mode = mode

        is_path = isinstance(filepath_or_object, (str, PurePath))
        exists = is_path and Path(filepath_or_object).exists()
        
        if is_path:
            self.filepath_or_object = smart_open(filepath_or_object, mode = self.mode)
        else:
            self.filepath_or_object = filepath_or_object

        if "r" in self.mode:
            self.read_header(header)
        elif "w" in self.mode or "a" in self.mode:
            self.header = header
            file_is_empty = ("a" in self.mode and self.filepath_or_object.seekable() and self.filepath_or_object.tell() == 0) or "w" in self.mode
            if file_is_empty:
                self.filepath_or_object.write(self.header.to_string())

    def read_header(self, header: PairsHeader = None):
        if isinstance(header, PairsHeader):
            self.header = header
            return
        file_like = hasattr(self.filepath_or_object, "seek") and hasattr(self.filepath_or_object, "readline")
        assert file_like, f"PairsFile filepath_or_object {self.filepath_or_object} must have tell, seek and readline methods"
        self.filepath_or_object.seek(0)

        lines = []
        while True:
            pos = self.filepath_or_object.tell()
            line = self.filepath_or_object.readline()
            if not line.startswith("#"):
                self.filepath_or_object.seek(pos)
                break
            else:
                lines.append(line)
        header_text = "".join(lines)
        self.header = PairsHeader.from_text(header_text)

    def pair_segment_from_text(self, line):
        import time
        
        # From here...
        stripped = line.strip()
        if not stripped:
            raise StopIteration
        fields = stripped.split()
        field_vals = {self.indexed_columns[idx]: val for idx, val in enumerate(fields)}
        # .. to here takes about 0.74 seconds

        
        # This statement takes about 2.3 seconds
        record = PairsSegment(**field_vals)
        
        return record

    def __iter__(self):
        self.indexed_columns = dict(enumerate(self.header.columns))
        return self

    def __next__(self):
        line = self.filepath_or_object.readline()
        record = self.pair_segment_from_text(line)
        return record


    def to_header(self):
        self.filepath_or_object.seek(0)
    
    def to_records(self, record_number=0):
        self.to_header()

        # First, scan for the first non-comment line
        while True:
            position = self.filepath_or_object.tell()  # Get the current file pointer position
            line = self.filepath_or_object.readline()  # Read a line manually

            if not line:  # If we reach EOF, exit the loop
                break

            if not line.startswith("#"):
                self.filepath_or_object.seek(position)  # Seek back to the start of the non-comment line
                if record_number == 0:
                    return line

        # If record_number is not 0, continue reading lines
        while record_number > 0:
            line = self.filepath_or_object.readline()
            record_number -= 1
            if not line:  # Handle the case where EOF is reached
                return None
            if record_number == 0:
                return line

    def write(self, pairs_segment: PairsSegment):
        self.filepath_or_object.write("\n" + pairs_segment.to_string())



from collections import Counter
from copy import deepcopy
from functools import reduce
from numbers import Number
from scipy.optimize import linprog
from statistics import mean
import numpy as np
import random
import math

# See tests/test_discrete_distribution.py for unit tests


class DiscreteDistribution(Counter):
    @classmethod
    def mean_mass(cls, distributions: list['DiscreteDistribution']) -> 'DiscreteDistribution':
        """Compute the mean probability mass over event space.

        Sums the values assigned to each event in the sample space, defaulting to zero
        if unspecified, and divides by the total to obtain mean probability mass of each event.
        """
        # If sum() is used, it attempts to add individual values from one DiscreteDistribution
        # to another DiscreteDistribution, which isn't defined, so this is used instead.
        combined = reduce(lambda a, b: a + b, distributions)
        return combined / combined.total()
    
    @classmethod
    def bounded_probabilistic_round(cls, n, n_min, n_max):
        """Round n up or down probabilitistically within bounds (n_min, n_max)
        """
        # Testing difficulties: stochastic
        lower = max(n_min, math.floor(n)) if n_min else math.floor(n)
        upper = min(n_max, math.ceil(n)) if n_max else math.ceil(n)
        prob_up = n - lower
        if random.random() < prob_up:
            return upper
        else:
            return lower
            

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_size(self, n: Number, lower_bounds = {}, upper_bounds = {}) -> 'DiscreteDistribution':
        """Downsample to a specific number of reads (n is integer) or multiple of reads (n is float)"""
        # Testing difficulties: behaves quite differently depending on n's type 
        total = n if isinstance(n, int) else self.total() * n

        new_counts = {}
        for event, prob in self.probabilities().items():
            lower_bound = lower_bounds.get(event, 0)
            upper_bound = upper_bounds.get(event, float('inf'))
            new_counts[event] = DiscreteDistribution.bounded_probabilistic_round(prob*total, lower_bound, upper_bound)
        
        return DiscreteDistribution(new_counts)
    
    @property
    def size(self) -> Number:
        return sum(self.values())

    def downsample_to_probabilities(self, probdist: 'DiscreteDistribution') -> 'DiscreteDistribution':
        """Given a probability distribution, minimally downsample counts to match it.

        Find the maximum N such that 0 <= N*probdist[event] <= self[event] for all events
        """
        # Testing difficulty: linear programming result would have to be hand-verified

        # Cost function (linear programming attempts to minimize cost, so we
        # aim to minimize the negative sum of counts in order to maximize
        # the sum of counts)
        c = [-1]

        # Upper bounds. A_ub is a one-column vector of the probabilities.
        # b_ub is a list of the original counts, serving as the upper bound.
        # Linear programming will ensure that [A_ub]N <= b_ub

        events = self.events()
        A_ub = np.array([probdist.outcomes(events)]).T
        b_ub = list(self.outcomes(events))

        # Set lower bound of 0 on N to force positive counts
        bounds = [(0, None)]

        # Find value of N meeting criteria
        result = linprog(c, A_ub=A_ub, b_ub=[b_ub], bounds=bounds)
        
        assert result.success, "No solution found attempting to match distribution"

        # Get value of N - total number of downsampled events
        N = DiscreteDistribution.bounded_probabilistic_round(result.x[0], 0, self.size)

        return probdist.to_size(N, lower_bounds = {event: 0 for event in events}, upper_bounds = self)

    def probabilities(self) -> "DiscreteDistribution":
        return self.copy() / self.total()

    def __lt__(self, other: 'DiscreteDistribution') -> bool:
        return self.total() < other.total()

    def __le__(self, other: 'DiscreteDistribution') -> bool:
        return self.total() <= other.total()

    def __gt__(self,other: 'DiscreteDistribution') -> bool:
        return self.total() > other.total()

    def __ge__(self, other: 'DiscreteDistribution') -> bool:
        return self.total() >= other.total()

    def __add__(self, other: 'DiscreteDistribution') -> 'DiscreteDistribution':
        support = set(self.events() + other.events())
        event_outcome = {event: self[event] + other[event] for event in support}
        return DiscreteDistribution(event_outcome)
    
    def __truediv__(self, denom: Number) -> 'DiscreteDistribution':
        result = self.copy()
        for event in self.events():
            result[event] /= denom
        return result

    def copy(self) -> 'DiscreteDistribution':
        return deepcopy(self)

    def outcomes(self, events = None) -> list[Number]:
        if not events:
            return list(self.values())
        else:
            return [self[event] for event in events]

    def events(self) -> list[object]:
        return list(self.keys())


from hich.cli.paramlist import ParamList

# See tests/test_cli_types.py for unit tests

class _StrList(ParamList):
    name = "str_list"

    def value_type(self):
        return str

    def to_type(self, value):
        try:
            return str(value)
        except:
            raise ValueError(f"{value} is not a valid string")

StrList = _StrList()


import click

@click.command
@click.option("--format", "--fmt", "fmt",
    type = click.Choice(['fasta', 'fastq', 'seqio', 'sam', 'bam', 'sambam', 'alignment', 'pairs']), required = True,
    help = "Alignment data format")
@click.option("--f1", "-f", "--file", "--file1", type = str, required = True,
    help = "Path to first (or only) input sequencing data file")
@click.option("--f2", "--file2", "f2", default = None, type = str, show_default = True,
    help = "Path to second input sequencing data file")
@click.option("--out-dir", "--dir", "--output-dir", type = str, default = "",
    help = "Output directory")
@click.option("--annot-file", "--annot", "-a", "--annotations", default = None, type = str,
    help = ("Path to annotation file, a columnar text file mapping data file "
            "information such as a cell barcode to new information such as "
            "an experimental condition or cell ID. Annotation files with headers "
            "convert to a dict with format {col1_row: {col2:col2_row, col3:col3_row...}}."))
@click.option("--annot-has-header", "-h", type = bool, default = False, show_default = True,
    help = "Whether or not annotation file has a header row.")
@click.option("--annot-separator", "-s", type = str, default = "\t", show_default = repr('\t'),
    help = "The column separator character")
@click.option("--head", "--n_records", type = int, default = None, show_default = True,
    help = "Take only the first n records from the data files. Takes all records if not provided.")
@click.option("--key-code", "--kc", type = str, default = None,
    help = "Python code to extract an annotation row key from the current record.")
@click.option("--record-code", "--rc", type = str, default = None,
    help = "Python code to modify the record")
@click.option("--output-code", "--fc", "--filename", type = str, default = None,
    help = "Python code to select the record output.")
def organize(fmt,
             f1,
             f2,
             out_dir,
             annot_file,
             annot_has_header,
             annot_separator,
             head,
             key_code,
             record_code,
             output_code):
    """Reannotate and split sequencing data files


    """
    raise NotImplementedError("Hich organize is not implemented yet")


import click
from pathlib import Path
from tinydb import TinyDB, Query
from json import loads
import datetime
import os
import pandas as pd
import sys

@click.command
@click.option('--report', '-r', type = click.Path(), default = 'hich_all_output.html', show_default=True, help = 'Path of output HTML report')
@click.option('--db', '-d', type = click.Path(), default = 'hich_all_output.json', show_default=True, help = 'Path of output JSON database')
@click.option('--gather', '-g', type = str, default = 'hich_output.json', show_default = True, help = 'Glob expression for filename to gather')
@click.option("--work", '-w', type = click.Path(), default = ".", show_default = True, help = 'Root of work directory to gather from')
@click.option('--if-exists', '-e', default = 'use-unchanged', type = click.Choice(['use-unchanged', 'overwrite', 'append'], case_sensitive = False), help = 'Behavior when database exists.')
@click.option('--silent', '-s', is_flag = True, default = False, help = "Report messages on conflicts and errors")
def gather(report, db, gather, work, if_exists, silent):
    def convert_to_absolute_paths(data, work: Path) -> dict:
        if isinstance(data, dict):
            # If data is a dictionary, apply function to each value
            return {key: convert_to_absolute_paths(value, work) for key, value in data.items()}
        elif isinstance(data, list):
            # If data is a list, apply function to each item in the list
            return [convert_to_absolute_paths(item, work) for item in data]
        elif isinstance(data, str):
            # If data is a string and it represents an existing path, convert it to absolute path
            path = work / data
            if path.exists():
                return str(path.resolve())
            else:
                return data

        return data
    
    db_path = Path(db)
    if db_path.exists() and if_exists == 'overwrite':
        db_path.unlink()
        if not silent:
            print(f"Overwriting database at {db_path}", file = sys.stderr)

    db = TinyDB(db_path)

    if (db_path.exists() and (if_exists == 'append' or if_exists == 'overwrite')) or not db_path.exists():
        for path in Path(work).rglob(gather):
            with open(path) as file:
                record = loads(file.read())
            date_modified = datetime.datetime.fromtimestamp(path.stat().st_mtime)
            metadata = {'metadata': {'stage_dir': str(path.parent.resolve()), 'modified': str(date_modified)}}
            record = convert_to_absolute_paths(record, path.parent)
            record.update(metadata)
            db.insert(record)
    elif if_exists == 'use-unchanged' and not silent:
        print(f"Using original database at {db_path}", file = sys.stderr)
    
    records = sorted(db.all(), key = lambda x: x['metadata']['modified'])
    df = pd.json_normalize(records)
    df.to_html(report)


    


import time

# See tests/test_pairs_segment.py for unit tests

class PairsSegment:
    reserved = {"readID": str, "chr1": str, "pos1": int,
                "chr2": str, "pos2": int, "strand1": str, "strand2": str}
    required = {"chr1": str, "pos1": int, "chr2": str, "pos2": int}
    alt = {"chrom1":"chr1", "chrom2":"chr2"}

    def __init__(self, **kwargs):
        self.__dict__ = kwargs
        self.cast_reserved()
        self.alt_to_main()
    
    def cast_reserved(self):
        self.pos1 = int(self.pos1) if "pos1" in self.__dict__ else 0
        self.pos2 = int(self.pos2) if "pos2" in self.__dict__ else 0
    
    def alt_to_main(self):
        for alt, main in PairsSegment.alt.items():
            # Access self.__dict__ directly for faster lookups
            alt_val = self.__dict__.get(alt)
            main_val = self.__dict__.get(main)

            if alt_val is None and main_val is not None:
                self.__dict__[alt] = main_val
            elif main_val is None and alt_val is not None:
                self.__dict__[main] = alt_val
    

    def to_dict(self, columns = None):
        if columns:
            return {c:self.__dict__[c] for c in columns if c in self.__dict__}
        else:
            reserved = {k:self.__dict__[k] for k in PairsSegment.reserved
                 if k in self.__dict__}
            non_reserved = {k:v for k, v in self.__dict__.items()
                    if k not in PairsSegment.alt
                    and k not in reserved}
            reserved.update(non_reserved)
            return reserved

    def to_string(self, columns = None):
        return "\t".join(str(v) for v in self.to_dict(columns).values())

    @property
    def distance(self):
        return abs(self.pos1 - self.pos2) if hasattr(self, 'chr1') and self.is_cis else None
    
    @property
    def meets_spec(self):
        return all([hasattr(self, requirement) for requirement in PairsSegment.required])

    @property
    def is_cis(self): return self.chr1 == self.chr2
    
    @property
    def is_trans(self): return self.chr1 != self.chr2

    @property
    def intrachr(self): return self.is_cis

    @property
    def interchr(self): return self.is_trans

    @property
    def is_ur(self): return self.pair_type in ["UU", "RU", "UR"]

    def __str__(self): return self.to_string()


import click
import numpy as np
import polars as pl
import time
import warnings
import sys
from hich.parse.pairs_parser import PairsParser
from hich.fragtag.frag_index import FragIndex
from hich.fragtag.bedpe_pairs import BedpePairs
from hich.fragtag.samheader_fragtag import SamheaderFragtag

def tag_restriction_fragments(frags_filename: str,
                              input_pairs_filename: str,
                              output_pairs_filename: str,
                              batch_size: int = 1000000):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    
    frag_index = FragIndex(frags_filename)
    pairs_parser = PairsParser(input_pairs_filename)

    for df in pairs_parser.batch_iter(batch_size):  
        df = BedpePairs(df).fragtag(frag_index)

        pairs_parser.write_append(output_pairs_filename,
                                  df,
                                  header_end = SamheaderFragtag())


    pairs_parser.close()




from hich.cli.paramlist import ParamList

# See tests/test_cli_types.py for unit tests

class _IntList(ParamList):
    name = "int_list"

    def value_type(self):
        return int

    def to_type(self, value):
        try:
            return int(value) if value else None
        except:
            raise ValueError(f"'{value}' is not a valid integer")

IntList = _IntList()


import click
from hich.stats import DiscreteDistribution, compute_pairs_stats_on_path, load_stats_and_classifier_from_file, aggregate_classifier
from typing import List
from pathlib import Path

@click.command
@click.option("--to-group-mean", is_flag = True, default = False)
@click.option("--to-group-min", is_flag = True, default = False)
@click.option("--to-size", type = str, default = None)
@click.option("--prefix", type = str, default = "aggregate_")
@click.option("--outlier", type = str, multiple=True)
@click.argument("stats-paths", type = str, nargs = -1)
def stats_aggregate(to_group_mean, to_group_min, to_size, prefix, outlier, stats_paths):
    """Aggregate hich stats files called over .pairs with same conjuncts
    """
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi

    # Load the stats files into dataframes.
    # Ensure the stats files have identical conjuncts.
    # If they have a stratum column, aggregate all the unique strata from each
    # stratum column to get the complete collection of strata.
    # Use the classifier to build DiscreteDistributions from the dataframes.
    classifier, distributions = aggregate_classifier(stats_paths)

    # Get the complete collection of distributions.
    targets = [d for d in distributions]

    build_prefix = ""
    if to_group_mean:
        # Get the mean probability mass for all distributions.
        non_outliers = [distribution for distribution, path in zip(distributions, stats_paths) if path not in outlier]
        group_mean = DiscreteDistribution.mean_mass(non_outliers)
        
        # Downsample each individual sample by the minimum amount necessary to match its mean probabilities for the group.
        targets = [d.downsample_to_probabilities(group_mean) for d in distributions]
        if prefix is None:
            build_prefix += "to_group_mean"
    if to_group_min:
        # Downsample all samples to the size of the smallest one.
        # Then downsample all samples to that size.
        min_size: int = min(targets).total()
        targets: List[DiscreteDistribution] = [d.to_size(min_size) for d in targets]

        if prefix is None:
            build_prefix += "to_group_min"
    if to_size:
        if to_size.isdigit():
            to_size = int(to_size)
        else:
            to_size = float(to_size)
        targets = [d.to_size(to_size) for d in targets]
        if prefix is None:
            build_prefix += f"to_{to_size}"
    if prefix is None:
        prefix = build_prefix + "_"
    for d, stats_path in zip(targets, stats_paths):
        df = classifier.to_polars(d)
        path = str(Path(stats_path).parent / (prefix + Path(stats_path).name))
        df.write_csv(path, separator = "\t", include_header = True)


import click
from hich.cli import IntList, StrList, BooleanList
from hich.hicrep_combos import hicrep_combos

@click.command
@click.option("--resolutions", type = IntList, default = 10000)
@click.option("--chroms", "--include_chroms", type = StrList, default = None)
@click.option("--exclude", "--exclude_chroms", type = StrList, default = None)
@click.option("--chrom-filter", type=str, default = "chrom if size > 5000000 else None")
@click.option("--h", type = IntList, default = "1")
@click.option("--d-bp-max", type = IntList, default = "-1")
@click.option("--b-downsample", type = BooleanList, default = False)
@click.option("--nproc", type=int, default=None)
@click.option("--output", type=str, default = None)
@click.argument("paths", type=str, nargs = -1)
def hicrep(resolutions, chroms, exclude, chrom_filter, h, d_bp_max, b_downsample, nproc, output, paths):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    result = hicrep_combos(resolutions, chroms, exclude, chrom_filter, h, d_bp_max, b_downsample, nproc, output, paths)
    if result is not None:
        click.echo(result)


from smart_open import smart_open
from typing import Union
import click
import io
import pandas as pd
import polars as pl
import smart_open
import smart_open_with_pbgzip
import sys
import warnings

def read_pairs(read_from, 
batch_size: int = 10000, 
yield_columns_line: bool = True, 
exception_on_4dn_violation: bool = True) -> Union[str, pl.DataFrame]:
    """
    Generator that reads 4DN .pairs format in batches.
    https://github.com/4dn-dcic/pairix/blob/master/pairs_format_specification.md

    The first yielded item is the header as a string. If yield_columns_line is True,
    then the complete header will be returned. Otherwise, any lines starting with
    #columns: will be excluded from the header. This is to facilitate changing
    the header if the pairs table is to be reshaped.

    Subsequent yielded items are Polars DataFrames containing up to
    batch_size .pairs records. The fields will be named according to the
    #columns: line, whether or not yield_columns_line is True or False. All
    types will be strings.
    """
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    # Used to accumulate header lines or records for a batch
    header_lines = []
    records = []
    columns = None

    # Iterate through all the lines in the iterable
    for i, line in enumerate(read_from):
        
        if exception_on_4dn_violation:
            # Optionally, raise an exception on 4DN spec violations

            # Enforce that file starts with '## pairs format v1.0'
            if header_lines and not header_lines[0].startswith("## pairs format v1.0"):
                raise Exception((
                    f"4DN spec violation found in {read_from}, line {i}: {line}\n"
                    f"first header line is {header_lines[0]} but must start "
                    "with '## pairs format v1.0'"
                ))
            if header_lines is None:
                # Enforce that all header lines come before the first data line
                if line.startswith("#"):
                    raise Exception(
                        (f"4DN spec violation found in {read_from}, line {i}: {line}\n"
                        "header line starting with '#' "
                        "appears on line {i}, after the first data line. All header "
                        "lines must come before data lines.")
                    )
        
        if line.startswith("#columns:"):
            # When a #columns: line is found, use the subsequent
            # whitespace-separated strings as column names, with String type
            # for the polars schema. Add the line to the header lines.
            columns = line.split()[1:]
            schema = {col: pl.String for col in columns}

            if yield_columns_line:
                # We give the option not to add #columns: lines into the header
                # because this lets the client conveniently reshape the yielded
                # DataFrames to have a different schema and write the #columns:
                # line based on the result.
                header_lines.append(line)
        elif line.startswith("#"):
            # Add other header lines
            header_lines.append(line)
        else:
            if exception_on_4dn_violation and columns is None:
                # Enforce that there is a #columns: line somewhere in the
                # .pairs file header. Note that per the spec, the #columns:
                # line does NOT have to go immediately before the data entries.
                raise Exception((
                    f"4DN spec violation found in {read_from}, line {i}: {line}\n"
                    "No #columns: line found prior to first data line."
                ))

            # We are in the data entries now, so split the entries into individual
            # fields and accumulate in the growing batch.
            records.append(line.split())

            if header_lines is not None:
                # header_lines is a list prior to finding the first data entry.
                # If we enter this block, it means we just found the first
                # data entry, meaning we need to build the complete header
                # string and yield it. Then we set header_lines to None to
                # mark that we've exited the header region.
                header = "".join(header_lines)
                header_lines = None
                yield header
            elif len(records) == batch_size:
                # We have reached the target batch size, so build a dataframe
                # with the accumulated records and yield it as the latest batch.
                # Then reset the records to start accumulating another batch.
                result = pl.DataFrame(records, orient='row', schema=schema)
                records = []
                yield result
    if records:
        # If we have accumulated a partial batch when we run out of records,
        # yield them in a final partial batch.
        yield pl.DataFrame(records, orient='row', schema=schema)

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi
class PairsParser:
    def __init__(self, filename):
        self.filename = filename
        self.write_file = None
        
    def columns_row(self):
        """Returns the 0-indexed offset of the final header row with the column names

        row_index: 0-indexed offset of the final header row (header rows start with '#')
        columns_dict: {colname:index} dict of which columns correspond to which column names

        If a row starting with '#columns:' is present it is assumed to be the final header row.
        If no '#columns:' row is present, returns the last header row with the minimal seven
        column names given by the .pairs specification.
        
        Returns: (row_index, columns_dict)
        """

        default_columns = ['readID', 'chrom1', 'pos1', 'chrom2', 'pos2', 'strand1', 'strand2']
        row = 0

        with smart_open.open(self.filename, "rt", encoding='utf-8') as file:
            for line in file:
                if line.startswith('#columns:'):
                    column_names = line.split()[1:]
                    return (row, column_names)
                elif line.startswith('#'):
                    row += 1
                else:
                    return (row-1, default_columns)
        return (row - 1, default_columns)

    def header(self):
        """Returns list of all lines starting with '#' until but not including the first line not starting with '#'
        """
        header = []
        with smart_open.open(self.filename, "rt", encoding="utf-8") as file:
            for line in file:
                if line.startswith('#'):
                    header.append(line)
                else:
                    break
        return header

    def batch_iter(self, n_rows):
        def read_chunk(skip_rows, columns):
            return pl.read_csv(self.filename,
                             skip_rows = skip_rows,
                             n_rows = n_rows,
                             raise_if_empty = False,
                             separator = '\t',
                             has_header = False,
                             new_columns = columns,
                             dtypes = {"readID":pl.String,
                                       "chrom1":pl.String,
                                       "chrom2":pl.String})
            
        columns_row, columns = self.columns_row()

        skip_rows = columns_row + 1
        with open(self.filename) as file:
            df = read_chunk(skip_rows, columns)
            skip_rows += len(df)
            yield df
            while len(df) == n_rows:
                df = read_chunk(skip_rows, columns)
                skip_rows += len(df)
                if not df.is_empty():
                    yield df

    def write_append(self, filename, df = None, header_end = None):
        warnings.filterwarnings("ignore", message="Polars found a filename")

        if self.write_file is None:
            self.write_file = smart_open.open(filename, "w")
            header_no_columns = self.header()[:-1]
            final_header_line_fields = header_no_columns[-1].split()
            
            if header_end:
                pn_field = [field[3:]
                    for field in final_header_line_fields
                    if field.startswith("PN:")]
                
                header_end.PP = pn_field[0] if pn_field else "null"
                
            header_no_columns = "".join(header_no_columns)
            columns = " ".join(["#columns:"] + df.columns) + "\n"
            header_end = str(header_end)
            header = header_no_columns + header_end + columns
            self.write_file.write(header)

        if df is not None:
            buffer = io.StringIO()
            df.write_csv(buffer,
                        include_header = False,
                        separator = '\t')
            buffer.seek(0)
            self.write_file.write(buffer.getvalue())
            buffer.close()

    def close(self):
        if self.write_file is not None:
            self.write_file.close()
            self.write_file = None


import polars as pl
"""
Annotation files are required to either have headers or must be given
headers at the command line.

A zero-offset row index will be included. It can be given a name by the user.
If it is not included, then any of the following names can be used to access it:
    "index", "idx", "Index"
These names only refer to index columns if it did not already exist in the annotation file.

A key column must be supplied. A particular annotation is accessed by extracting it
from information in the data file using annotations[key].index or annotations[key][index]

I think the way this should work is that we extract the key from the record object, obtaining
a value of 'key'

"""

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

class AnnotationFile(dict):

    def read_csv(self, file, key_col = None, **kwargs):
        df = pl.read_csv(file, **kwargs)
        key_col = key_col or df.columns[0]
        for i, row in enumerate(self.iter_rows(named = True)):
            key = row[key_col]
            self[key] = {"index": i, "annotations": row}
    
    def __getitem__(self, key):
        if key not in self:
            self.update({key: {"index": len(self)}})
        return self.get(key)



from hich.cli.paramlist import ParamList
from pathlib import Path

# Not currently unit tested as of 2024/10/20 - Ben Skubi

class _PathList(ParamList):
    name = "path_list"

    def value_type(self):
        return Path

    def to_type(self, value):
        try:
            return Path(value)
        except:
            raise ValueError(f"{value} is not a valid path")

PathList = _PathList()


import polars as pl
import pysam
import Bio
from hich.parse.seqio_splitter import SeqIOSplitter
from hich.parse.alignment_splitter import AlignmentSplitter
from hich.parse.pairs_file import PairsFile
from hich.parse.pairs_segment import PairsSegment
from hich.parse.pairs_splitter import PairsSplitter
from hich.parse.annotation import AnnotationFile
from pathlib import Path
from collections import defaultdict

def organize(fmt: str,
             f1_path: str,
             f2_path: str = None,
             out_dir: str = None,
             annot_file: str = None,
             annot_has_header: str = False,
             annot_separator: str = "\t",
             head: int = None,
             key_code: str = None,
             record_code: str = None,
             output_code: str = None):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    def yield_none():
        while True: yield None

    parsers = {
        ("pairs",): {"file": PairsFile, "splitter": PairsSplitter},
        ("fastq", "fasta", "seqio"): {"file": Bio.SeqIO.parse, "splitter": SeqIOSplitter},
        ("sam", "bam", "sambam", "alignment"): {"file": pysam.AlignmentFile, "splitter": AlignmentSplitter}
    }
    parser = None
    for parser_key in parsers:
        if fmt in parser_key: parser = parsers[parser_key]
    
    file_type = parser["file"]
    splitter = parser["splitter"]()
    
    f1 = file_type(f1_path)
    f2 = file_type(f2_path) if f2_path else yield_none()
    
    if isinstance(splitter, PairsSplitter) or isinstance(splitter, AlignmentSplitter):
        splitter.header = f1.header

    files = zip(f1, f2)
    if key_code: key_code = compile(key_code, '<string>', 'eval')
    if record_code: record_code = compile(record_code, '<string>', 'exec')
    if output_code: output_code = compile(output_code, '<string>', 'eval')
    annotations = AnnotationFile()
    if annot_file:
        annotations.read_csv(annot_file, has_header = annot_has_header, separator = annot_separator)
    for i, records in enumerate(files):
        record = records[0]
        record1, record2 = records
        key = eval(key_code) if key_code else None
        annotation = annotations[key]
        
        if record_code: exec(record_code)
        output = eval(output_code)
        if output:
            splitter.write(output, record1)
            if record2: splitter.write(output, record2)
        if i >= head:
            break




from hich.cli.paramlist import ParamList

# See tests/test_cli_types.py for unit tests

class _BooleanList(ParamList):
    name = "boolean_list"

    def value_type(self):
        return bool

    def to_type(self, value):
        if value.lower() in ('true', '1', 'yes'):
            return True
        elif value.lower() in ('false', '0', 'no'):
            return False
        else:
            raise ValueError(f"'{value}' is not a valid boolean")

BooleanList = _BooleanList()



import click
from hich.fragtag import tag_restriction_fragments

@click.command
@click.option("--batch_size", default = 1000000)
@click.argument("fragfile")
@click.argument("out_pairs")
@click.argument("in_pairs")
def fragtag(batch_size, fragfile, out_pairs, in_pairs):
    tag_restriction_fragments(fragfile, in_pairs, out_pairs, batch_size)


from hich.cli.booleanlist import BooleanList, _BooleanList
from hich.cli.intlist import IntList, _IntList
from hich.cli.pathlist import PathList, _PathList
from hich.cli.strlist import StrList, _StrList
from hich.cli.paramlist import ParamList

__all__ = ['BooleanList', '_BooleanList', 'IntList', '_IntList', 'PathList', '_PathList', 'StrList', '_StrList', 'ParamList']


import numpy as np
import polars as pl

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

class FragIndex:
    def __init__(self, filename = None):
        if filename:
            self.load_bed(filename)

    def load_bed(self, filename):
        bedcols = ['chrom', 'start', 'end']
        self.chrom_intervals = pl.read_csv(filename,
                                            separator = '\t',
                                            has_header = False,
                                            new_columns = bedcols) \
                                 .sort(by=['chrom', 'start', 'end']) \
                                 .partition_by(['chrom'],
                                               as_dict = True,
                                               include_key = False)

        self.chrom_ends_sorted = {chrom:
                                  sorted(self.ends(chrom).to_list())
                                  for chrom
                                  in self.chrom_intervals}
    
    def search(self, chrom, positions):
        return np.searchsorted(self.chrom_ends_sorted[chrom], positions)
    
    def starts(self, chrom):
        return self.chrom_intervals[chrom]['start']
    
    def ends(self, chrom):
        return self.chrom_intervals[chrom]['end']
    
    def __contains__(self, chrom):
        return chrom in self.chrom_intervals


from hicrep.hicrep import sccByDiag
from hicrep.utils import *
import scipy.sparse as sp

def cool2pixels(cool: cooler.api.Cooler):
    """Return the contact matrix in "pixels" format

    Args:
        cool: Input cooler object

    Returns:
        cooler.core.RangeSelector2D object
    """
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    return cool.matrix(as_pixels=True, balance=True, sparse=True)



def hicrepSCC(cool1: cooler.api.Cooler, cool2: cooler.api.Cooler,
              h: int, dBPMax: int, bDownSample: bool,
              chrNames: list = None, excludeChr: set = None) -> dict:
    """Compute hicrep score between two input Cooler contact matrices

    Args:
        cool1: `cooler.api.Cooler` Input Cooler contact matrix 1
        cool2: `cooler.api.Cooler` Input Cooler contact matrix 2
        h: `int` Half-size of the mean filter used to smooth the
        input matrics
        dBPMax `int` Only include contacts that are at most this genomic
        distance (bp) away
        bDownSample: `bool` Down sample the input with more contacts
        to the same number of contacts as in the other input
        chrNames: `list` List of chromosome names whose SCC to
        compute. Default to None, which means all chromosomes in the
        genome are used to compute SCC
        excludeChr: `set` Set of chromosome names to exclude from SCC
        computation. Default to None.

    Returns:
        `float` scc scores for each chromosome
    """
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    
    # Get binsize for each contact matrix
    binSize1 = cool1.binsize
    binSize2 = cool2.binsize

    # Input validation:
    # Same bin sizes and number of bins
    # Same number of chroms
    # Same chrom names
    
    assert binSize1 == binSize2,\
        f"Input cool files have different bin sizes"
    assert coolerInfo(cool1, 'nbins') == coolerInfo(cool2, 'nbins'),\
        f"Input cool files have different number of bins"
    assert coolerInfo(cool1, 'nchroms') == coolerInfo(cool2, 'nchroms'),\
        f"Input cool files have different number of chromosomes"
    assert (cool1.chroms()[:] == cool2.chroms()[:]).all()[0],\
        f"Input file have different chromosome names"
    
    
    binSize = binSize1
    bins1 = cool1.bins()
    bins2 = cool2.bins()
    
    # Use median bin size if bin size is non-uniform
    if binSize is None:
        # sometimes bin size can be None, e.g., input cool file has
        # non-uniform size bins.
        assert np.all(bins1[:] == bins2[:]),\
            f"Input cooler files don't have a unique bin size most likely "\
            f"because non-uniform bin size was used and the bins are defined "\
            f"differently for the two input cooler files"
        # In that case, use the median bin size
        binSize = int(np.median((bins1[:]["end"] - bins1[:]["start"]).values))
        warnings.warn(f"Input cooler files don't have a unique bin size most "\
                      f"likely because non-uniform bin size was used. HicRep "\
                      f"will use median bin size from the first cooler file "\
                      f"to determine maximal diagonal index to include", RuntimeWarning)
    
    # Validate dBPMax
    
    if dBPMax == -1:
        # this is the exclusive upper bound
        dMax = coolerInfo(cool1, 'nbins')
    else:
        dMax = dBPMax // binSize + 1

    assert dMax > 1, f"Input dBPmax is smaller than binSize"
    
    # Get pixels
    p1 = cool2pixels(cool1)
    p2 = cool2pixels(cool2)
    
    # get the total number of contacts as normalizing constant
    n1 = coolerInfo(cool1, 'sum')
    n2 = coolerInfo(cool2, 'sum')

    # Use dict here so that the chrNames don't duplicate
    if chrNames is None:
        chrNamesDict = dict.fromkeys(cool1.chroms()[:]['name'].tolist())
    else:
        chrNamesDict = dict.fromkeys(chrNames)
    
    # It's important to preserve the order of the input chrNames so that the
    # user knows the order of the output SCC scores so we bail when encounter
    # duplicate names rather than implicit prunning the names.
    assert chrNames is None or len(chrNamesDict) == len(chrNames), f"""
        Found Duplicates in {chrNames}. Please remove them.
        """
    # filter out excluded chromosomes
    if excludeChr is None:
        excludeChr = set()

    chrNames = [ chrName for chrName in chrNamesDict if chrName not in excludeChr ]

    # Create scc score vector with placeholder value
    scc = np.full(len(chrNames), -2.0)

    # Compute scc scores
    for iChr, chrName in enumerate(chrNames):
        # normalize by total number of contacts
        mS1 = getSubCoo(p1, bins1, chrName)
        
        # Input validation on shape of chromosome matrix
        assert mS1.size > 0, "Contact matrix 1 of chromosome %s is empty" % (chrName)
        assert mS1.shape[0] == mS1.shape[1],\
            "Contact matrix 1 of chromosome %s is not square" % (chrName)
        
        mS2 = getSubCoo(p2, bins2, chrName)
        assert mS2.size > 0, "Contact matrix 2 of chromosome %s is empty" % (chrName)
        assert mS2.shape[0] == mS2.shape[1],\
            "Contact matrix 2 of chromosome %s is not square" % (chrName)
        assert mS1.shape == mS2.shape,\
            "Contact matrices of chromosome %s have different input shape" % (chrName)
        try:
            scc[iChr] = computeSCC(mS1, mS2, n1, n2, dMax, h, bDownSample)
        except Exception as e:
            print(e)

    return scc

def computeSCC(mS1: sp.coo_matrix,
               mS2: sp.coo_matrix,
               n1: float,
               n2: float,
               dMax: int,
               h: float,
               bDownSample: bool) -> float:
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    # Compute scc score
    nDiags = mS1.shape[0] if dMax < 0 else min(dMax, mS1.shape[0])
    rho = np.full(nDiags, np.nan)
    ws = np.full(nDiags, np.nan)
    # remove major diagonal and all the diagonals >= nDiags
    # to save computation time
    m1 = trimDiags(mS1, nDiags, False)
    m2 = trimDiags(mS2, nDiags, False)
    del mS1
    del mS2
    if bDownSample:
        # do downsampling
        size1 = m1.sum()
        size2 = m2.sum()
        if size1 > size2:
            m1 = resample(m1, size2).astype(float)
        elif size2 > size1:
            m2 = resample(m2, size1).astype(float)
    else:
        # just normalize by total contacts
        m1 = m1.astype(float) / n1
        m2 = m2.astype(float) / n2
    if h > 0:
        # apply smoothing
        m1 = meanFilterSparse(m1, h)
        m2 = meanFilterSparse(m2, h)

    return sccByDiag(m1, m2, nDiags)


import click
from hich.stats import DiscreteDistribution
from hich.pairs import PairsClassifier, PairsFile, PairsSegment
from hich.cli import IntList, StrList
from hich.io import df_to_disk_or_stdout

def count_pairs_stats(classifier: PairsClassifier, pairs_file: PairsFile) -> DiscreteDistribution:
    """Get a DiscreteDistribution of counts of pairs in the pairs_file classified as events"""

    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    distribution = DiscreteDistribution()

    for record in pairs_file:
        outcome = classifier.classify(record)
        distribution[outcome] += 1
    return distribution

@click.command
@click.option("--conjuncts",
    type = StrList,
    default = "chrom1, chrom2, pair_type, stratum",
    show_default = True,
    help = "PairsSegment traits that define the category for each record (comma-separated string list)")
@click.option("--cis-strata",
    type = IntList,
    default = "",
    show_default = True,
    help = "PairsSegment cis distance strata boundaries for use with 'stratum' conjunct (comma-separated string list)")
@click.option("--output",
    type = click.Path(writable=True),
    default = "",
    show_default = True,
    help = "Output file for tab-separated stats file. If not given, outputs to stdout.")
@click.argument("pairs", type = click.Path(exists=True, dir_okay=False))
def stats(conjuncts: str, cis_strata: IntList, output: click.Path, pairs: click.Path) -> None:
    """
    Classify pairs and count the events.

    Output has conjuncts as headers, one row per event, and a column "count" containing the count of each event.

    Can read 4DN .pairs format from plaintext or from a variety of compressed formats with Python's smart_open package.

    Example:
        hich stats --conjuncts "chr1 chr2" --cis-strata "10000 20000" my_pairs_file.pairs.gz
    """
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi

    classifier = PairsClassifier(conjuncts, cis_strata)
    pairs_file = PairsFile(pairs)
    distribution = count_pairs_stats(classifier, pairs_file)
    df = classifier.to_polars(distribution)
    df_to_disk_or_stdout(df, output, include_header=True, separator = "\t")


from multiprocessing import Pool
from typing import Tuple
import polars as pl
from polars import DataFrame
from pathlib import Path

def compute_pairs_stats_on_path(data: Tuple["PairsClassifier", Path]) -> Tuple[str, "DiscreteDistribution"]:
    """Classify records in a PairsFile as events and return their counts
    data - a (PairsClassifier, Path) tuple for classifying PairsSegments from a PairsFile as events
    and counting the number of events.
    """
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    from hich.stats import DiscreteDistribution
    from hich.pairs import PairsClassifier, PairsFile, PairsSegment

    classifier, pairs_path = data
    pairs_file = PairsFile(pairs_path)
    stats = DiscreteDistribution()

    # Return a count of events in the pairs file
    for record in pairs_file:        
        outcome = classifier.classify(record)
        stats[outcome] += 1
    result = (pairs_path, stats)
    return result

def aggregate_classifier(pairs_stats_paths: list[str]) -> Tuple["PairsClassifier", list["DiscreteDistribution"]]:
    """Determine the support of the combined pairs stats files

    pairs_stats_paths -- list of filenames containing pairs stats
    """
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    from hich.pairs import PairsClassifier
    from hich.stats import DiscreteDistribution
    dfs = []
    conjuncts = None
    raw_strata = set()

    # Collect dataframes of all input stats files
    for path in pairs_stats_paths:
        df = pl.read_csv(path, separator = "\t", infer_schema_length = None)
        dfs.append(df)

        # Ensure conjuncts match for all stats files.
        new_conjuncts = [col for col in df.columns if col and col != "count"]
        conjuncts = conjuncts or new_conjuncts
        assert conjuncts == new_conjuncts, "Conjuncts do not match for all hich stats files."

        # Get unique strata listed in "stratum" column if present
        if "stratum" in conjuncts:
            strata = set(s for s in df["stratum"].unique().to_list() if s is not None)

    # Sort strata from lowest to highest; extract to integer or float.
    raw_strata = sorted(list(raw_strata))
    cis_strata = []
    for stratum in raw_strata:
        if stratum.isdigit():
            cis_strata.append(int(stratum))
        else:
            try:
                cis_strata.append(float(stratum))
            except ValueError:
                pass

    # Build a classifier using the conjuncts and complete collection of cis strata
    # Then use the classifier to convert all the dataframes into DiscreteDistributions
    classifier = PairsClassifier(conjuncts, cis_strata)
    distributions = []
    for df in dfs:
        distributions.append(classifier.from_polars(df))
    return classifier, distributions



def load_stats_and_classifier_from_file(pairs_stats_header_tsv_path):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    from hich.stats import DiscreteDistribution
    from hich.pairs import PairsClassifier, PairsFile
    df = pl.read_csv(pairs_stats_header_tsv_path, separator = "\t", infer_schema_length = None)
    conjuncts = [col for col in df.columns if col != "count"]
    cis_strata = None
    if "stratum" in conjuncts:
        raw_strata = df["stratum"].unique().to_list()
        cis_strata = []
        for s in raw_strata:
            if str(s).isdigit():
                cis_strata.append(int(s))
            elif s is not None:
                try:
                    cis_strata.append(float(str(s)))
                except ValueError:
                    print("Not integer or float", s)
                    cis_strata.append(s)
    classifier = PairsClassifier(conjuncts, cis_strata)
    distribution = DiscreteDistribution()
    for row in df.iter_rows(named = True):
        count = row["count"]
        outcome = tuple([row[conjunct] for conjunct in conjuncts])
        distribution[outcome] = count
    return classifier, distribution

def compute_pairs_stats_on_path_list(classifier, pairs_paths):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    data = [(classifier, path) for path in pairs_paths]
    return Pool().map(compute_pairs_stats_on_path, data)


from dataclasses import dataclass
from smart_open import smart_open
from hich.parse.pairs_header import PairsHeader
from hich.parse.pairs_segment import PairsSegment
from hich.parse.pairs_file import PairsFile
from hich.parse.file_splitter import FileSplitter

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class PairsSplitter(FileSplitter):
    """ Write to 4DN Consortium .pairs-format files, creating handles as needed
    """
    header: PairsHeader = None

    def access(self, filename: str):
        if filename not in self.handles:
            self.handles[filename] = PairsFile(filename, "w", header = self.header)
        return self.handles[filename]
    
    def write(self, filename: str, pairs_segment: PairsSegment):
        self.access(filename).write(pairs_segment)




import cooler
from typing import Protocol, Dict, Iterable, Union, List, Any
import pandas
from pathlib import Path
from dataclasses import dataclass, field
import sys

class BinPixelParser(Protocol):
    """Construct inputs to Open2C cooler create_scool method
    """

    @classmethod
    def can_parse(cls, obj) -> bool:
       """Return whether the IterDictParser can parse the object"""
       ...

    @classmethod
    def cell_name(cls, obj: Any, config: Dict) -> str:
        """Return the name associated with the object"""
        ...

    @classmethod
    def cell_name_bins_dict(
        cls, 
        cell_name: str,
        obj: Any,
        config: Dict,
        shared_bins: Union[None, pandas.DataFrame] = None,
        extended_bins_cols: Union[bool, List[str]] = True,
        *args,
        **kwargs
        ) -> Dict[str, pandas.DataFrame]:
        """Return a {cell_name (str): bins (pandas.DataFrame)} dict
        
        When inserting large batches of cells, some aspects of cooler.create_scool should be kept in mind.

        1. It raises an exception unless bins is a pandas.DataFrame containing the columns "chrom", "start" and "end". However, it does not use these columns unless working on the first item in the entire bins dict.

        1a. In the case where only these three columns are used, the same instance of the bins dataframe can be passed for every cell. This can be created once and specified using shared_bins, in which case typical behavior for a BinPixelParser is to assign that bins dataframe instance as the value for every value in the returned dict.
        
        1b. When passing bins dataframes containing extended columns (i.e. weights), the unnecessary "chrom", "start" and "end" columns must be included, along with the cell-specific bin-associated columns we actually want to write. All of these must be loaded into memory at once for all cells to be written, which could be memory-intensive if the number of cells is large. To mitigate this, typical behavior for the BinPixelParser is to set the value of the "chrom", "start" and "end" columns to None in the bins DataFrame, which should conserve memory consumption when passing in a large number of these DataFrames, but still allow extended columns to be written. Users may wish to filter the extended columns to be written to a subset, in which case they can pass a list of the columns that should be included.
        
        If the shared_bins DataFrame was not given to the ScoolCreator object, then the chrom, start, and end columns of the first bins DataFrame created by BinPixelParser will be used as the shared_bins. It should be sorted by (chrom, start, end).
        """
        ...

    @classmethod
    def cell_name_pixels_dict(
        cls, 
        cell_name: str,
        obj: Union[cooler.Cooler, str],
        config: Dict,
        shared_bins: pandas.DataFrame,
        extended_pixels_cols: Union[bool, List[str]] = True,
        regions_iter: Iterable = None,         
        *args,
        **kwargs
        ) -> Dict[str, Iterable]:
        """Return a {cell_name (str): pixels (Iterable[pandas.DataFrame])} dict
        
        Cooler.create_scool, when receiving a dictionary of this format as its chrom_names_pixels_dict, will iterate through the values for each cell to be written to the .scool file. It expects each yielded object to be a pandas DataFrame containing the columns bin1_id, bin2_id, and count, possibly along with extra extended columns. These columns will be written iteratively for that cell. The bin1_id and bin2_id columns correspond to the 0-indexed offset of a value in the chrom, start, and end datasets and of the corresponding values in the chrom, start, and end row of the shared_bins (when created).

        Arguments:
            cell_name: str: The name of the cell to write the pixels to
            obj: Union[cooler.Cooler, str]: The Cooler object or .cool filename
            config: Dict: Details of how to process the object
            shared_bins: pandas.DataFrame: Shared bins DataFrame, potentially useful to create the pixels DataFrame in some cases
            extended_pixels_cols: Union[bool, List[str]]: If a True boolean, keep all loaded pixels columns. If False, keep only chrom1, start1, end1, chrom2, start2, end2, count. If a list of column names, keep specified subset of extended columns in addition to chrom1, start1, end1, chrom2, start2, end2, count.
            regions_iter (Iterable): Yields pandas.DataFrames covering all pixels to be written for the cell
        """
        ...

class CoolBinPixelParser:
    """Parses cooler.Cooler objects or .cool files"""

    no_cooler_cell_name = ValueError(
"""
Could not determine cell name for Cooler object or file.
Options to associate a Cooler object or file with a cell name include:

1. In the config parameter, pass {"cell": "desired cell name"}
2. In the config parameter, pass {"Cooler.info_cell_key": "key"} where "key" is the key in the Cooler.info metadata dict containing the desired cell name
3. Pass a filename such that Path(filename).name returns a string
"""
)
    @classmethod
    def can_parse(cls, obj) -> bool:
        """Return whether the CoolBinPixelParser can parse the object"""
        if isinstance(obj, cooler.Cooler):
            return True
        elif isinstance(obj, str):
            try:
                cooler.Cooler(obj)
                return True
            except:
                return False
        return False

    @classmethod
    def cell_name(cls, obj: Union[cooler.Cooler, str] = None, config: Dict = {}) -> str:
        """Return the name associated with the object"""
        # Various ways of getting the cell name
        if "cell" in config:
            return config["cell"]
        elif isinstance(obj, str):
            path = Path(obj)
            return str(path.name.removesuffix(path.suffix))
        elif isinstance(obj, cooler.Cooler) and "Cooler.info_cell_key" in config:
            return str(obj.info[config["Cooler.info_cell_key"]])
        else:
            raise CoolBinPixelParser.no_cooler_cell_name

    @classmethod
    def cell_name_bins_dict(
        cls, 
        cell_name: str,
        obj: Union[cooler.Cooler, str],
        config: Dict,
        shared_bins: Union[None, pandas.DataFrame] = None,
        extended_bins_cols: Union[bool, List[str]] = True,
        *args,
        **kwargs
        ) -> Dict[str, pandas.DataFrame]:
        """Return cell bin dict for a cooler.Cooler object or filename to a .cool file
        """
        if not isinstance(cell_name, str):
            raise TypeError(f"cell_name should be str, is {type(cell_name)}")
        # If a filename was passed, create the Cooler object
        if isinstance(obj, str):
            obj = cooler.Cooler(obj)
        # 1. If we received shared bins and we don't want to use extended bins cols, then use the shared bins
        # 2. If we didn't receive shared bins, or if extended bins cols are to be kept, then load bins from the file.
        #   2a. If we received shared bins, but extended bins cols are kept, then set the chrom, start, end to placeholders so we don't pass redundant info to create_scool
        #   2b. If we did not receive shared bins, and extended bin cols are kept, then return the shared cols (chrom, start, end) plus whatever extended bin cols should be kept
        #   2b_1. If we received a list of extended bin col names, use the subset present in the Cooler object
        #   2b_2. If we received a True value for extended bin cols, use the entire loaded bins dataframe

        if shared_bins and isinstance(extended_bins_cols, bool) and not extended_bins_cols:
            # Condition 1
            # Use shared bins if we don't need extended bins columns (i.e. "weight") and
            # shared bins have been passed
            bins = shared_bins
        else:
            # Condition 2
            # We don't have shared bins or we need extended columns
            # Either way we need to load the bins from the cool file.
 
            # Create combined bins dict for all chromosomes, which will be used
            # as-is if shared_bins hasn't been passed and we want to keep the extended columns
            bins = pandas.concat([
                obj.bins().fetch(chromname)
                for chromname in obj.chromnames
            ])

            if shared_bins:
                # Condition 2a
                # cooler.create explicitly validates that the bins is a pandas DataFrame
                # that contains chrom, start, and end columns. However, it doesn't actually use
                # those columns for writing per-cell bin information. In cases where per-cell bins
                # information is being passed, we therefore replace them with a small placeholder
                # to avoid filling memory with a large amount of unnecessary information.
                bins["chrom"] = None
                bins["start"] = None
                bins["end"] = None
            else:
                # Condition 2b
                if isinstance(extended_bins_cols, list):
                    # Condition 2b.1
                    # Get subset of columns present in bins
                    cols = ['chrom', 'start', 'end'] + extended_bins_cols
                    cols = [col for col in cols if col in bins.columns]
                elif isinstance(extended_bins_cols, bool):
                    # Condition 2b.2
                    if extended_bins_cols:
                        cols = bins.columns
                    else:
                        cols = ['chrom', 'start', 'end']
                else:
                    raise ValueError(f"extended_bins_cols should be bool or list of column names, was {type(extended_bins_cols)}")
                bins = bins[cols]
                
                
        # Return the cell_name_bins_dict
        return {cell_name: bins}

    @classmethod
    def cell_name_pixels_dict(
        cls,
        cell_name: str,
        obj: Union[cooler.Cooler, str],
        config: Dict,
        shared_bins: pandas.DataFrame,
        extended_pixels_cols: Union[bool, List[str]] = True,
        regions_iter: Iterable = None,
        *args,
        **kwargs
        ) -> Dict[str, Iterable]:
        """Return cell pixel iter dict for a cooler.Cooler object or filename to a .cool file
        """
        # If received a filename, create Cooler object
        if isinstance(obj, str):
            obj = cooler.Cooler(obj)
        
        # Declare object as unambiguously Cooler
        obj: cooler.Cooler

        CoolBinPixelParser.validate_bin_size(obj, config)

        # Iterate through chromnames by default
        regions_iter = regions_iter or obj.chromnames

        # Create function to yield pandas DataFrames
        def load_regions(c: cooler.Cooler, extended_pixels_cols: List[str], regions_iter: Iterable):
            for region in regions_iter:
                pixels = c.pixels().fetch(region)
                base_cols = ["chrom1", "start1", "end1", "chrom2", "start2", "end2", "count"]
                if isinstance(extended_pixels_cols, bool):
                    if not extended_pixels_cols:
                        pixels = pixels[base_cols]
                elif isinstance(extended_pixels_cols, list):
                    pixels = pixels[base_cols + extended_pixels_cols]
                yield pixels
        
        # Create the pixel iter dict
        return {cell_name: load_regions(obj, extended_pixels_cols, regions_iter)}

    @staticmethod
    def validate_bin_size(c: cooler.Cooler, config):
        # Validate that if a target resolution is specified, the cooler matches
        target_resolution = config.get('resolution', None)
        if target_resolution is not None and c.binsize != target_resolution:
            raise ValueError(f"Cooler.binsize is {c.binsize} but config['resolution'] is {target_resolution}. Set to None for no filter or the true resolution/binsize of this Cooler object.")

class ScoolCreator:
    """Creates inputs and calls cooler.create_scool

    Allows registering parsers for specific objects and file formats
    """
    # Registered parsers for ScoolCreator objects
    parsers: List[BinPixelParser] = None

    @dataclass
    class InputObject:
        """Specifies an input object to parse and details on how to parse it"""
        # Object to parse, which will only be processed if at least one BinPixelParser's .can_parse() method returns True for it
        obj: Any

        # Parser-specific configuration flags
        config: Dict = field(default_factory=dict)

        # Columns other than the base columns 'chrom', 'start', and 'end' to write for the specific cell
        # If True writes all pixels columns, if False writes only the base columns
        extended_bins_cols: Union[bool, List[str], pandas.DataFrame] = True
        
        # Columns other than the base columns 'chrom1', 'start1', 'end1', 'chrom2', 'start2', 'end2', 'count' to write
        # If True writes all pixels columns, if False writes only the base columns
        extended_pixels_cols: Union[bool, List[str], pandas.DataFrame] = True

        # Iterable or generator that returns pixels dataframes in chunks on-the-fly while cooler.create_scool iterates through it
        # Useful to conserve memory by loading pixels in batches from disk as they are being written.
        regions_iter: Iterable[pandas.DataFrame] = None

    def __init__(self):
        self.parsers = ScoolCreator.parsers.copy()


    @classmethod
    def register(cls, parser: BinPixelParser) -> None:
        """Register a parser as the highest-priority parser"""
        if cls.parsers is None:
            cls.parsers = []
        cls.parsers = [parser] + cls.parsers


    def create_scool(self, cool_uri: str, input_objects: List[InputObject], shared_bins: pandas.DataFrame = None, **kwargs):
        """Extracts bins and pixels from specified objects and calls cooler.create_scool to create/append to an .scool file at the given cool_uri

        kwargs are passed to the cooler.create_scool method.

        input_objects: a list of objects to extract from. Currently supported by hich-cli:
            filenames of .cool files (str)
            cooler.Cooler objects (cooler.Cooler)
        

        Additional classes following the BinPixelParser Protocol can be registered with the ScoolCreator class or added to the list of parsers for an ScoolCreator instance to extend compatibility to new objects, file formats, etc.

        Note that if called in append mode (mode = 'a'), then cooler.create_scool will delete and recreate the shared bins group. However, .scool is an HDF5-based data model and in HDF5 and deleting HDF5 groups leaves dead space in the file. If called many times, this could result in a substantially bloated file. This dead space can be removed by installing hdf5-tools and calling h5repack on the file.
        """
        # Will be updated with cell names as keys, bins DataFrames as values
        cell_name_bins_dict = {}

        # Will be updated with cell names as keys, pixel DataFrame iterators/generators or DataFrames
        cell_name_pixels_dict = {}

        for obj in input_objects:
            # Coerce to an input object
            if not isinstance(obj, ScoolCreator.InputObject):
                obj = ScoolCreator.InputObject(obj)
            
            # Get the BinPixelParser capable of parsing this object
            parser: BinPixelParser = self.select_parser(obj.obj)

            if parser:
                # If found, then get the name, bins and pixels and update the bins and pixels dicts
                print(f"Preparing bins and pixels insertion dict for {obj}", file = sys.stderr)
                cell_name = parser.cell_name(obj.obj, obj.config)
                cell_bins = parser.cell_name_bins_dict(cell_name, obj.obj, obj.config, shared_bins, obj.extended_bins_cols)
                shared_bins = shared_bins or cell_bins
                cell_pixels = parser.cell_name_pixels_dict(cell_name, obj.obj, obj.config, shared_bins, obj.extended_pixels_cols, obj.regions_iter)

                cell_name_bins_dict.update(cell_bins)
                cell_name_pixels_dict.update(cell_pixels)
            else:
                print(f"Skipping {obj} as no registered parser was found", file = sys.stderr)

        print(f"All input objects that could be parsed are ready for insertion. Creating {cool_uri}.", file = sys.stderr)
        # Create the .scool file
        cooler.create_scool(
            cool_uri = cool_uri,
            bins = cell_name_bins_dict,
            cell_name_pixels_dict = cell_name_pixels_dict,
            **kwargs
        )
        if Path(cool_uri).exists():
            print(f"{cool_uri} created successfully.", file = sys.stderr)

    def select_parser(self, obj: InputObject) -> Union[None, BinPixelParser]:
        """Select the first (highest-priority) parser that can parse the object"""
        for parser in self.parsers:
            if parser.can_parse(obj):
                return parser
        return None

# Register hich-cli default BinPixelParsers with the ScoolCreator class
ScoolCreator.register(CoolBinPixelParser)



import click
from hich.digest import make_fragment_index

@click.command()
@click.option("--output", default = None, show_default = True, help = "Output file. Compression autodetected by file extension. If None, prints to stdout.")
@click.option("--startshift", default = 0, show_default = True, help = "Fixed distance to shift start of each fragment")
@click.option("--endshift", default = 0, show_default = True, help = "Fixed distance to shift end of each fragment")
@click.option("--cutshift", default = 1, show_default = True, help = "Fixed distance to shift cutsites")
@click.argument("reference")
@click.argument("digest", nargs = -1)
def digest(output, startshift, endshift, cutshift, reference, digest):
    """
    In silico digestion of a FASTA format reference genome into a
    BED format fragment index.

    Allows more than 800 restriction enzymes (all that are supported by
    Biopython's Restriction module, which draws on REBASE).

    Digest can also be specified as a kit name. Currently supported kits:

        "Arima Genome-Wide HiC+" or "Arima" -> DpnII, HinfI
    
    Multiple kits and a mix of kits and enzymes can be added. Duplicate
    kits, enzyme names, and restriction fragments are dropped.

    Can read compressed inputs using decompression autodetected and supported
    by the Python smart_open library. Can compress output using compression
    formats autodetected by the Polars write_csv function.

    Format is:
    chrom start cut_1
    chrom cut_1 cut_2

    The startshift param is added to all values in column 1.
    The endshift param is added to all values in column 2.
    """
    # We aim to support specification of digests by kit name
    # (potentially versioned), so this converts the kit names to the enzymes
    # used in that kit.
    make_fragment_index(output, startshift, endshift, cutshift, reference, digest)


import seaborn as sns
import time
import dash
from dash import dcc, html, callback, Output, Input, State, ALL, callback_context
import base64
import io
import polars as pl
from natsort import natsorted
import plotly.graph_objects as go
import seaborn as sns
import matplotlib.pyplot as plt
import time
import webbrowser
from threading import Timer

def seaborn_clustermap(symmetric_df, settings = {}):
    cell_size = .6
    cell_count = len(symmetric_df)
    size = min(16, cell_size*cell_count)
    figsize=(size, size)
    cell_size = size / cell_count
    sns.set(font_scale=cell_size/.6)  # Seaborn uses a relative scaling factor
    
    g = sns.clustermap(symmetric_df,
                        annot=True,
                        fmt=".2f",
                        cmap="viridis",
                        cbar=False,
                        dendrogram_ratio=(.05, .05),
                        figsize = figsize)
    g.ax_heatmap.set_facecolor("gray")
    g.cax.set_visible(False)
    
    return g

def scc_with_distance(data):
    # Compute distance metric from SCC score
    scc = data["scc"]
    dist = ((.5*(1 - scc))**.5).alias("score")
    
    # Rename SCC column to "score"
    scc_data = data.rename({"scc":"score"})
    dist_data = scc_data.clone()
    
    # Add distance in "score" column for dist_data
    dist_data = dist_data.with_columns(dist)

    # Create label for type of score
    scc_label = pl.Series(["scc"]*len(scc)).alias("cluster by")
    dist_label = pl.Series(["distance: 0 ≤ √[.5(1-scc)] ≤ 1"]*len(dist)).alias("cluster by")

    # Add score type label to column
    scc_data = scc_data.with_columns(scc_label)
    dist_data = dist_data.with_columns(dist_label)

    # Concatenate SCC and distance dataframes
    return pl.concat([scc_data, dist_data])


app = dash.Dash(__name__)

model = None
states = None
filename = ""
state = {}
dummy = {"__dummy__":[]}
options = {"__dummy__":[]}
multi_options = {"__dummy__":[]}
static_options = {"__dummy__":[]}
data_cols = ["file1", "file2", "score"]
cluster_data = None

# Placeholder for storing dataset options
app.layout = html.Div([
    dcc.Upload(id='upload-data', children=html.Button('Upload Data')),
    dcc.Store(id='__dummy__'),
    dcc.Store(id='multi-options'),
    dcc.Store(id='static-options'),
    html.Div(
        id = "clustermap-image",
        style={
            'display': 'flex',
            'justify-content': 'center',
            'align-items': 'center',
            'width': '100%'  # Ensure the container takes up full width
        }
    ),
    html.Div(id='static-options-container'),
    html.Div(id='slider-container')  # This will hold the sliders
])

@callback(
    [Output('multi-options', 'data'),
     Output('static-options', 'data')],
    Input('upload-data', 'contents'),
    State('upload-data', 'filename')
)
def load_data(contents, upload_filename):
    global model, data_cols, options, states, multi_options, static_options, dummy, filename
    if contents is None:
        return {}, {}

    content_type, content_string = contents.split(',')
    decoded = base64.b64decode(content_string)
    filename = upload_filename
    
    # Use io.BytesIO to create a file-like object from the decoded content
    uploaded = pl.read_csv(io.BytesIO(decoded), separator="\t")
    model = scc_with_distance(uploaded)

    options = {col: natsorted(model[col].unique())
               for col in model.columns
               if col not in data_cols}
    
    states = model.select(options.keys()).unique()
    state = states.row(0, named = True)
    
    multi_options = {col: options[col]
                     for col in options
                     if len(options[col]) > 1}
    static_options = {"filename": [filename]}
    static_options.update({col: options[col]
                     for col in options
                     if len(options[col]) == 1})

    multi_options.update(dummy)
    static_options.update(dummy)

    return multi_options, static_options

@callback(
    Output('slider-container', 'children'),
    Input('multi-options', 'data')
)
def update_multi_options(new_multi_options):
    global model, data_cols, options, states, multi_options, static_options, dummy
    if new_multi_options: new_multi_options.pop('__dummy__')
    if not new_multi_options:
        return []

    slider_elements = []
    for option_name, option_settings in new_multi_options.items():
        char_count = sum([len(str(setting)) for setting in option_settings])
        width = char_count * 9
        slider_elements.append(
            html.Div([
                dcc.Slider(id = {'type': 'dynamic-slider', 'index': option_name},
                           min = 0,
                           max = len(option_settings)-1,
                           step=None,
                           value = 0,
                           marks = {str(i): str(setting) for i, setting in enumerate(option_settings)})
            ], style={"width": f"{width}px", 'margin': '10px 0', 'justify-content': 'center'})
        )
    return html.Div(slider_elements, style={'display': 'flex', 'flex-direction': 'column', 'align-items': 'center', 'width': '100%'})


@callback(
    Output('static-options-container', 'children'),
    Input('static-options', 'data')
)
def update_static_options(new_static_options):
    global model, data_cols, options, states, multi_options, static_options, dummy
    if new_static_options: new_static_options.pop('__dummy__')
    if not new_static_options:
        return []

    p_elements = []
    for option_name, option_settings in new_static_options.items():
        p_elements.append(
            html.Div([html.P(f"{option_name}: {option_settings[0]}")])
        )

    return html.Div(p_elements, style={'display': 'flex', 'flex-direction': 'column', 'align-items': 'center', 'width': '100%'})

@app.callback(
    Output('clustermap-image', 'children'),
    [Input({'type': 'dynamic-slider', 'index': ALL}, 'value')]
)
def update_output(*slider_values):
    global model, data_cols, options, states, multi_options, static_options, dummy
    # Access callback context to get both IDs and values
    ctx = callback_context
    if not ctx.triggered or model.is_empty():
        return html.Div()

    # Extract the IDs and their corresponding values
    triggered_inputs = ctx.inputs_list[0]
    update = {}
    for trigger in triggered_inputs:
        option = trigger['id']['index']
        setting_idx = trigger['value']
        setting = options[option][setting_idx]
        update[option] = setting
    state.update(update)

    cluster_data = model.join(pl.DataFrame(state), on = state.keys()).select(data_cols)

    fig = clustermap_figure(cluster_data)
    return dcc.Graph(figure=fig)

def clustermap_figure(data):
    
    df = data.to_pandas()

    # Step 1: Pivot the DataFrame
    pivot_df = df.pivot(index='file1', columns='file2', values='score')
    
    # Step 2: Make the pivot symmetric by copying transposed values
    # Fill missing values in the pivot by using values from its transpose
    symmetric_df = pivot_df.combine_first(pivot_df.T)

    
    setting_labels = [f"filename: {filename}"] + [f"{option}: {setting}" for option, setting in state.items()]
    title = "HicRep Clustermap: " + ", ".join(setting_labels)
    settings = {
        "title": title
    }

    clustermap = seaborn_clustermap(symmetric_df, settings)

    # Save the clustermap as an image in memory
    buf = io.BytesIO()
    clustermap.savefig(buf, format="png")
    plt.close(clustermap.fig)
    buf.seek(0)
    

    # Encode the image in base64
    encoded_image = base64.b64encode(buf.getvalue()).decode()
    # Get the image dimensions for proper sizing
    width, height = clustermap.fig.get_size_inches() * clustermap.fig.dpi

    # Create and return the figure containing the updated image
    fig = go.Figure()

    fig.add_layout_image(
        dict(
            source=f"data:image/png;base64,{encoded_image}",
            xref="x",
            yref="y",
            x=0,
            y=height,
            sizex=width,
            sizey=height,
            sizing="stretch",
            opacity=1,
            layer="below"
        )
    )

    # Hide axes, gridlines, etc.
    fig.update_xaxes(visible=False, range=[0, width])
    fig.update_yaxes(visible=False, range=[0, height])
    fig.update_layout(
        margin=dict(l=0, r=0, t=0, b=0),
        xaxis=dict(showgrid=False, zeroline=False, fixedrange=True),
        yaxis=dict(showgrid=False, zeroline=False, fixedrange=True),
        width=width,
        height=height
    )
    

    return fig

def make_demo_file():
    from itertools import combinations, combinations_with_replacement, product, chain
    from random import random
    from collections import defaultdict
    files = [f"file{i}" for i in range(50)]
    chroms = [f"chr{i}" for i in range(22)] + ["chrX", "chrY", "chrM"]
    h = ["1"]
    dBPMax = [None]
    resolutions = [100000, 1000000]
    file_combos = combinations_with_replacement(files, 2)
    param_combos = product(chroms, resolutions, h, dBPMax)
    rows = [files + params for files, params in product(file_combos, param_combos)]
    columns = ["file1", "file2", "chrom", "resolution", "h", "dBPMax", "scc"]
    settings = defaultdict(list)
    for row in rows:
        for val, column in zip(row, columns):
            settings[column].append(val)
        settings["scc"].append(random())
    settings = pl.DataFrame(settings)
    settings.write_csv("demo.tsv", separator = "\t")

def open_browser(host, port):
    url = f"http://{host}:{port}/"
    webbrowser.open_new(url)

def run_dashboard(host, port):
    Timer(1, open_browser, args=(host, port)).start()
    app.run_server(host = host, port = port, debug=True, use_reloader=False)

if __name__ == '__main__':
    run_dashboard()




import click
import sys
from pathlib import Path
import polars as pl
import pandas as pd
from typing import NewType, Union, IO

ReadableToDF = NewType('ReadableToDF', Union[str, Path, IO[str], IO[bytes], bytes])
WriteableToDF = NewType('WriteableToDF', Union[str, Path, IO[str], IO[bytes], None])

def df_to_disk_or_stdout(df, output: Union[WriteableToDF, None], *args, **kwargs):
    "Write a DataFrame to csv or stdout if output_path is None"
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    output_file = output or sys.stdout
    if isinstance(df, pd.DataFrame):
        to_stdout = df.to_csv(output_file, *args, **kwargs)
    elif isinstance(df, pl.DataFrame):
        to_stdout = df.write_csv(output_file, *args, **kwargs)
    else:
        TypeError(f"df_to_disk_or_stdout expects polars.DataFrame or pandas.DataFrame, got {type(df)}")
    if isinstance(to_stdout, str):
        click.echo(to_stdout)


from abc import ABC, abstractmethod
from dataclasses import dataclass, field

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

@dataclass
class FileSplitter:
    handles: dict = field(default_factory = dict[str, object])
    remove: str = ""
    
    @abstractmethod
    def access(self, filename: str): ...
    """Get a file handle ready for writing records"""

    @abstractmethod
    def write(self, filename: str, record: object): ...
    """Write a record"""

    def __del__(self):
        """Remove a given file after the FileSplitter is destroyed"""
        if self.remove and Path(self.remove).exists():
            Path(self.remove).unlink()


from pathlib import Path
from parse import parse

# !Warning: this class has no specific unit test as of 2024/10/20 - Ben Skubi

class PairsHeader:
    version_prefix = "## pairs format v"

    def __init__(self, text = "", chromsizes = {}, columns = [], command = []):
        self.text = text
        self.chromsizes = chromsizes
        self.columns = columns
        self.command = command

    def to_dict(self):
        return self.__dict__
    
    def to_string(self):
        return self.text

    def set_columns(self, columns: list[str]):
        new_columns_line = "#columns: " + " ".join(columns) + "\n"
        self.columns = columns
        header = parse("{start}#columns: {columns}\n{end}", self.text) or parse("{start}#columns: {columns}", self.text)
        start = header["start"]
        end = header["end"] if "end" in header else ""
        self.text = start + new_columns_line + end

    @classmethod
    def from_text(self, from_text):
        header = PairsHeader()
        header.text = from_text
        lines = header.text.split("\n")
        assert lines[0].startswith(PairsHeader.version_prefix), f"Pairs must start with ## pairs format v1.0 but first line was {line}"
        header.pairs_format_version = lines[0].removeprefix("## pairs format v")
        for line in lines[1:]:
            if not line.startswith("#"):
                break
            fields = line.split()
            field_type = fields[0]
            rest = line.removeprefix(field_type).lstrip()
            if field_type == "#chromsize:":
                contig, size = fields[1:]
                header.chromsizes[contig] = size
            elif field_type == "#command:":
                header.command.append(rest)
            elif field_type == "#columns:":
                header.columns = fields[1:]
            elif field_type.endswith(":"):
                field_name = field_type[1:-1]
                if field_name not in self.__dict__:
                    setattr(header, field_name, rest)
                elif isinstance(header.__dict__[field_name], str):
                    header.__dict__[field_name] = [header.__dict__[field_name], rest]
                else:
                    header.__dict__[field_name].append(rest)
        return header
    
    def __repr__(self):
        return repr(str(self))

    def __str__(self):
        return self.to_string()


from hich.commands.compartments import compartments
from hich.commands.create import create_scool
from hich.commands.digest import digest
from hich.commands.downsample import downsample
from hich.commands.fragtag import fragtag
from hich.commands.gather import gather
from hich.commands.hicrep_comparisons import hicrep_comparisons
from hich.commands.hicrep import hicrep
from hich.commands.organize import organize
from hich.commands.reshape import reshape
from hich.commands.stats_aggregate import stats_aggregate
from hich.commands.stats import stats





__all__ = ['compartments', 'create_scool', 'digest', 'downsample', 'fragtag', 'gather', 'hicrep_comparisons', 'hicrep', 'organize', 'reshape', 'stats_aggregate', 'stats']


class PairsSchema:
    
    reserved = {"readID": str, "chr1": str, "pos1": int,
                "chr2": str, "pos2": int, "strand1": str, "strand2": str}
    required = {"chr1": str, "pos1": int, "chr2": str, "pos2": int}
    synonyms = {"chrom1": "chr1", "chrom2": "chr2"}

    @classmethod
    def make_official(cls, column_name):
        return PairsSchema.synonyms.get(column_name, column_name)

    @classmethod
    def required_colnames(cls):
        return list(PairsSchema.required.keys())
    
    @classmethod
    def reserved_colnames(cls):
        return list(PairsSchema.reserved.keys())

    @classmethod
    def common_unofficial_colnames(cls):
        return list(PairsSchema.synonyms.keys())


import click
from typing import List, Dict
import itertools
from hich.scool import ScoolCreator
from glob import glob
import duckdb as dd
import sys

@click.command
@click.option("-g", "--glob", "globs", default = [], show_default=True, multiple=True, help="File glob of input files to include, can include multiple times")
@click.option("-m", "--mode", "mode", default="w", type = click.Choice(["w", "a"]), show_default=True, help="Overwrite (w) or append (a) to .scool file")
@click.option("-c", "--config", "config_df", default=None, show_default=True, help="Columnar file containing config parameters on a per-input file basis")
@click.option("-b", "--bins", "bins", default=None, show_default=True, type = click.Path(readable=True, dir_okay=False), help="Columnar file containing the headers chrom, start, end; can be created automatically or extracted from some input file types such as .cool/.mcool/.scool")
@click.argument("scool_uri", type = click.Path(writable=True, dir_okay=False))
@click.argument("input_obj", nargs = -1, type = click.Path(writable=True, dir_okay=False))
def create_scool(globs: List[List[str]], mode: str, config_df: str, bins: click.Path, scool_uri: click.Path, input_obj: List[str]) -> None:
    """Create an Open2C format .scool file (a single-resolution, multi-cell file)

    The only currently supported input file type is .cool.
        For .cool files, cell names if not otherwise given are specified from the filename [CELL].cool as the [CELL] prefix
    
    If a columnar config dict is specified (-c, --config), it should have headers. The schema is sniffed by DuckDB. The following headers control insertion as follows:
        filename (required): The input file that config settings for the row will apply to
        cell: The group under which the data for the file will be stored in the .scool file, in [scool_uri]::/cells/[cell]
        resolution: Validates the data is or can be binned at that resolution, and uses it
        extended_bins_cols: A whitespace-separated list of extended bins columns (besides chrom, start, end) to retain for the file
            Note: all bins cols will be used if header not provided; if header provided and cell is blank, no extended cols will be used for the file
        extended_pixels_cols: A whitespace-separated list of extended bins columns (besides chrom, start, end) to retain for the file
            Note: all pixels cols will be used if header not provided; if header provided and cell is blank, no extended cols will be used for the file
    Arguments:
        scool_uri (path): The filepath under which the scool file should be created.
        input_obj (list of paths/glob): The filepaths to insert into the scool file
    """
    print("Getting list of input files", file = sys.stderr)
    # Create a list of input object paths
    input_objects = list(itertools.chain.from_iterable([glob(g) for g in globs])) + list(input_obj)

    # Create ScoolCreator.InputObjects from them that can be configured from the config df
    input_objects = {obj: ScoolCreator.InputObject(obj) for obj in input_objects}


    # Configure the created InputObjects
    if config_df:
        print(f"Parsing configuration file {config_df}", file = sys.stderr)
        config_df = dd.read_csv(config_df).pl()
        for row in config_df.iter_rows(named=True):
            row: Dict[str, Any]
            filename = row.get("filename")
            if filename not in input_objects:
                input_objects[filename] = ScoolCreator.InputObject(filename)
            if not filename:
                break
            extended_bins_cols = row.get("extended_bins_cols", None)
            input_objects[filename].extended_bins_cols = extended_bins_cols.split() if extended_bins_cols else []

            extended_pixels_cols = row.get("extended_bins_cols", None)
            input_objects[filename].extended_bins_cols = extended_pixels_cols.split() if extended_pixels_cols else []
            if "resolution" in row:
                input_objects[filename].config.update({"resolution": row["resolution"]})
            if "cell" in row:
                input_objects[filename].config.update({"cell": row["cell"]})

    # Convert back to a list of InputObjects
    input_objects = list(input_objects.values())


    

    # Load shared bins
    print(f"Loading bins from {bins}", file = sys.stderr)
    shared_bins = dd.read_csv(bins) if bins else None
    scool_creator = ScoolCreator()

    print(f"Preparing to insert {len(input_objects)} objects", file = sys.stderr)
    scool_creator.create_scool(
        scool_uri,
        input_objects,
        shared_bins,
        mode=mode
    )


from hich.stats.discrete_distribution import DiscreteDistribution
from hich.stats.pairs_classifier import PairsClassifier
from hich.stats.compute_pairs_stats import compute_pairs_stats_on_path_list, \
    compute_pairs_stats_on_path, aggregate_classifier, load_stats_and_classifier_from_file

__all__ = ['DiscreteDistribution', 'PairsClassifier', 'compute_pairs_stats_on_path_list', 
'compute_pairs_stats_on_path', 'aggregate_classifier', 'load_stats_and_classifier_from_file']


from hich.stats import DiscreteDistribution, PairsClassifier, compute_pairs_stats_on_path, load_stats_and_classifier_from_file, aggregate_classifier
from hich.pairs import PairsFile
from hich.sample import SelectionSampler
import click

@click.command
@click.option("--conjuncts",
    type = str,
    default = "record.chr1 record.chr2 record.pair_type stratum",
    show_default = True,
    help = "PairsSegment traits that define the category for each record (space-separated string list)")
@click.option("--cis-strata",
    type = str,
    default = "10 20 50 100 200 500 1000 2000 5000 10000 20000 50000 100000 200000 500000 1000000 2000000 5000000",
    show_default = True,
    help = "PairsSegment cis distance strata boundaries (space-separated string list)")
@click.option("--orig-stats",
    type = str,
    default = "",
    show_default = True,
    help = ("Stats file containing original count distribution. Can be produced with hich stats. "
            "Computed from conjuncts and cis_strata if not supplied. Overrides default conjuncts and cis_strata if they are supplied."))
@click.option("--target-stats",
    type = str,
    default = "",
    show_default = True,
    help = "Stats file containing target count distribution.")
@click.option("--to-size",
    type = str,
    default = "",
    show_default = True,
    help = ("Float on [0.0, 1.0] for fraction of records to sample, or positive integer number of counts to sample. "
            "If a target stats file is supplied, further downsamples it to the given count."))
@click.argument("input_pairs_path", type = str)
@click.argument("output_pairs_path", type = str)
def downsample(conjuncts, cis_strata, orig_stats, target_stats, to_size, input_pairs_path, output_pairs_path):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    orig_classifier, orig_distribution = load_stats_and_classifier_from_file(orig_stats) if orig_stats else (None, None)
    target_classifier, target_distribution = load_stats_and_classifier_from_file(target_stats) if target_stats else (None, None)
    
    if orig_classifier and target_classifier:
        assert orig_classifier.conjuncts == target_classifier.conjuncts, f"Original and target conjuncts do not match for {orig_stats} and {target_stats}."
        conjuncts = orig_classifier.conjuncts
        if "stratum" in orig_classifier.conjuncts and "stratum" in target_classifier.conjuncts:
            cis_strata = list(set(orig_classifier.cis_strata + target_classifier.cis_strata))        
    elif orig_classifier:
        conjuncts = orig_classifier.conjuncts
    elif target_classifier:
        conjuncts = target_classifier.conjuncts
    
    classifier = PairsClassifier(conjuncts, cis_strata)
    
    if to_size.isdigit():
        to_size = int(to_size)
    else:
        try:
            to_size = float(to_size)
        except ValueError:
            to_size = None
    
    if not orig_distribution:
        _, orig_distribution = compute_pairs_stats_on_path((classifier, pairs_path))
    if not target_distribution:
        assert to_size is not None, "No target distribution or count supplied for downsampling."
        target_distribution = orig_distribution.to_size(to_size)
    if to_size:
        target_distribution = target_distribution.to_size(to_size)
    sampler = SelectionSampler(full = orig_distribution, target = target_distribution)
    input_pairs_file = PairsFile(input_pairs_path)
    output_pairs_file = PairsFile(output_pairs_path, mode = "w", header = input_pairs_file.header)

    for record in input_pairs_file:
        outcome = classifier.classify(record)
        if sampler.sample(outcome):
            output_pairs_file.write(record)


import click
from pathlib import Path
from hich.cli import StrList
from hich.compartments import write_compartment_scores

@click.command
@click.option("--chroms", type = StrList, default = None)
@click.option("--exclude-chroms", type = StrList, default = None)
@click.option("--keep-chroms-when", type = str, default = None)
@click.option("--n_eigs", type = int, default = 1)
@click.argument("reference")
@click.argument("matrix")
@click.argument("resolution", type = int)
def compartments(chroms, exclude_chroms, keep_chroms_when, n_eigs, reference, matrix, resolution):
    # !Warning: this method has no specific unit test as of 2024/10/20 - Ben Skubi
    
    matrix = Path(matrix)
    reference = Path(reference)
    final_suffix = matrix.suffixes[-1]
    prefix = matrix.name.rstrip(final_suffix)

    write_compartment_scores(prefix, matrix, reference, resolution, chroms, exclude_chroms, keep_chroms_when, n_eigs)